"""Wrapper for SNAP Connect to allow futures to be returned

See the readme for usage examples
"""
import logging
import random
from collections import defaultdict, OrderedDict
from datetime import timedelta

from tornado.concurrent import Future
from tornado.gen import coroutine, Return, with_timeout, TimeoutError

import paho.mqtt.client as mqtt

from .snap_future import SnapFuture
from .packet_overhead import get_dmcast_packet_overhead, get_max_addrs
from .responses import MultiResult

DEFAULT_MCAST_TTL = 5
DEFAULT_MCAST_GROUP = 0x0001
DEFAULT_DMCAST_DELAY = 50  # Delay in milliseconds between dmcast responses

RPC_RETRIES = 3  # Number of times to attempt to send an RPC.  This is NOT related to retries from not hearing responses.

log = logging.getLogger('snap2mqtt-future-client')
log_rpc = logging.getLogger("snap2mqtt-future-client.rpc")  # lower level debugger

# Message type enumerations
MESSAGE_TYPE_RPC = "uRPC"
MESSAGE_TYPE_MCAST = "mRPC"
MESSAGE_TYPE_DMCAST = "dmRPC"
MESSAGE_TYPE_DATA = "uData"
MESSAGE_TYPE_MDATA = "mData"


class FutureClient(object):
    """Based on ideas found here:

    https://git.synapse-wireless.com/core/future-snap-connect
    """

    def __init__(self, client):
        """Wrap a BaseClient instance with the ability to return Futures

        :param snap2mqtt_client.client.BaseClient client: BaseClient instance to wrap.
        """
        self._client = client

        self._client.on_publish(self._on_publish)

        self._unique = defaultdict(int)  # Store the next unique number for each function callback

        self._app_seq = random.randint(1, 255)  # Application level sequence number

        self._rpc_futures = {}  # RPC Sent ID -> Future
        self._callback_futures = {}  # Address.FunctionName -> Future

        self._immediately_resolve_rpcs = False
        self.immediately_resolve_rpcs(False)

        self.next_message_sent = Future()  # Future that resolves as soon as the next message sent

    def immediately_resolve_rpcs(self, immediately_resolve_rpcs=True):
        """Set S2MFC to immediately resolve RPC futures without on_publish being called first.

        This is mostly useful during testing.
        """
        self._immediately_resolve_rpcs = immediately_resolve_rpcs

    def _on_publish(self, client, userdata, mid):
        try:
            self._rpc_futures[mid].set_result(True)
            del self._rpc_futures[mid]
            self._message_sent()
        except KeyError:
            log_rpc.info("Could not find an existing Future for Message id %s", mid)
            log_rpc.debug("Outstanding RPC futures: %r", self._rpc_futures.keys())

    @staticmethod
    def get_default_mcast_ttl():
        """Return the default mutlicast TTL

        Override in a base class if your application changes this dynamically
        """
        return DEFAULT_MCAST_TTL

    @staticmethod
    def get_default_mcast_group():
        """Return the default mutlicast group

        Override in a base class if your application changes this dynamically
        """
        return DEFAULT_MCAST_GROUP

    @staticmethod
    def get_default_dmcast_delay():
        """Return the default directed multicast delay

        Override in a base class if your application changes this dynamically
        """
        return DEFAULT_DMCAST_DELAY

    def _message_sent(self):
        """Called every time a message is successfully sent"""
        self.next_message_sent.set_result(True)
        self.next_message_sent = Future()

    def _get_unique_name(self, name):
        """Return a unique version of the given name"""
        self._unique[name] += 1
        return "%s%s" % (name, self._unique[name])

    def get_next_seq(self):
        """Get the next application sequence number"""
        self._app_seq = (self._app_seq + 1) % 256

        if self._app_seq == 0:
            self._app_seq = 1

        return self._app_seq

    def uRPC(self, addr, funcname, *args):
        """Return a future that resolves when an RPC is sent

        :param str addr: SNAP Address (Ex: "aabbcc")
        :param str funcname: Function to call
        :param tuple *args: Arguments for RPC
        """
        future = Future()
        log_rpc.info("RPC: %s.%s%r", addr, funcname, args)

        result, mid = self._client.send({b'type': b'uRPC',
                                         b'dst': addr,
                                         b'func': funcname,
                                         b'args': args})

        log_rpc.info("Sent Ucast RPC: %r (ID: %r)", result, mid)

        if self._immediately_resolve_rpcs:
            future.set_result(result == mqtt.MQTT_ERR_SUCCESS)
            self._message_sent()
        else:
            if result == mqtt.MQTT_ERR_SUCCESS:
                self._rpc_futures[mid] = future
            else:
                # Failed already, no need to wait for on_publish
                future.set_result(False)

        return future

    def mRPC(self, group, ttl, funcname, *args, **kwargs):
        """Return a future that resolves when an mcast RPC is sent

        :param str group: Multicast group
        :param int ttl: Number of hops
        :param str funcname: Function to call
        :param tuple *args: Arguments for RPC
        """
        future = Future()

        log_rpc.info("Mcast RPC: %s%r", funcname, args)
        result, mid = self._client.send({b'type': b'mRPC',
                                         b'group': group,
                                         b'ttl': ttl,
                                         b'func': funcname,
                                         b'args': args})

        log_rpc.info("Sent Mcast RPC: %r (ID: %r)", result, mid)

        if self._immediately_resolve_rpcs:
            future.set_result(result == mqtt.MQTT_ERR_SUCCESS)
            self._message_sent()
        else:
            if result == mqtt.MQTT_ERR_SUCCESS:
                self._rpc_futures[mid] = future
            else:
                # Failed already, no need to wait for on_publish
                future.set_result(False)

        return future

    def dmRPC(self, addrs, group, ttl, delay, funcname, *args, **kwargs):
        """Return a future that resolves when an mcast RPC is sent

        :param list[str] addrs: List of SNAP Address (Ex: ["aabbcc", "ccddee"])
        :param str group: Multicast group
        :param int ttl: Number of hops
        :param int delay: millisecond delay between dmcast responses
        :param str funcname: Function to call
        :param tuple *args: Arguments for RPC
        """
        future = Future()

        log_rpc.info("DMcast RPC: %s%r", funcname, args)

        result, mid = self._client.send({b'type': b'dmRPC',
                                         b'dst': addrs,
                                         b'group': group,
                                         b'ttl': ttl,
                                         b'delay': delay,
                                         b'func': funcname,
                                         b'args': args})

        log_rpc.info("Sent DMcast RPC: %r (ID: %r)", result, mid)

        if self._immediately_resolve_rpcs:
            future.set_result(result == mqtt.MQTT_ERR_SUCCESS)
            self._message_sent()
        else:
            if result == mqtt.MQTT_ERR_SUCCESS:
                self._rpc_futures[mid] = future
            else:
                # Failed already, no need to wait for on_publish
                future.set_result(False)

        return future

    def uData(self, addr, data):
        """Return a future that resolves when the Data packet is sent

        :param str addr: SNAP Address (Ex: "aabbcc")
        :param str data: Data to send
        """
        future = Future()
        log_rpc.info("Data: %s.%r", addr, data)

        result, mid = self._client.send({b'type': b'uData',
                                         b'dst': addr,
                                         b'data': data})

        log_rpc.info("Sent Ucast Data: %r (ID: %r)", result, mid)

        if self._immediately_resolve_rpcs:
            future.set_result(result == mqtt.MQTT_ERR_SUCCESS)
            self._message_sent()
        else:
            if result == mqtt.MQTT_ERR_SUCCESS:
                self._rpc_futures[mid] = future
            else:
                # Failed already, no need to wait for on_publish
                future.set_result(False)

        return future

    def mData(self, group, ttl, data):
        """Return a future that resolves when an mcast RPC is sent

        :param str group: Multicast group
        :param int ttl: Number of hops
        :param str data: Data to send
        """
        future = Future()

        log_rpc.info("Mcast Data: %r", data)
        result, mid = self._client.send({b'type': b'mData',
                                         b'group': group,
                                         b'ttl': ttl,
                                         b'data': data})

        log_rpc.info("Sent Mcast Data: %r (ID: %r)", result, mid)

        if self._immediately_resolve_rpcs:
            future.set_result(result == mqtt.MQTT_ERR_SUCCESS)
            self._message_sent()
        else:
            if result == mqtt.MQTT_ERR_SUCCESS:
                self._rpc_futures[mid] = future
            else:
                # Failed already, no need to wait for on_publish
                future.set_result(False)

        return future

    def traceroute(self, addr):
        """Return a future that resolves when an mcast RPC is sent

        :param str addr: SNAP Address (Ex: "aabbcc")
        """
        future = Future()

        log_rpc.info("Traceroute: %s", addr)

        result, mid = self._client.send({b'type': b'traceroute',
                                         b'dst': addr})

        if self._immediately_resolve_rpcs:
            future.set_result(result == mqtt.MQTT_ERR_SUCCESS)
            self._message_sent()
        else:
            if result == mqtt.MQTT_ERR_SUCCESS:
                self._rpc_futures[mid] = future
            else:
                # Failed already, no need to wait for on_publish
                future.set_result(False)

        return future

    def subscribe_rpc(self, func_name, callback):
        """Subscribe to func_name RPCs with the callback.

        RPCs of any type will trigger the callback.

        Callback signature should be f(source_addr, *args).
        """
        self._client.subscribe('rpc/+/' + func_name, callback)

    def unsubscribe_rpc(self, func_name, callback):
        """Unsubscribe the RPC callback"""
        self._client.unsubscribe('rpc/+/' + func_name, callback)

    def unsubscribe_all(self, func_name):
        """Unsubscribe all callbacks for the given RPC"""
        # Matches SNAP Connect behavior...
        self._client.subscriptions.pop('rpc/+/' + func_name)

    def _get_callback(self, funcname):
        """Return a future callback for the given function name

        :rtype: callable
        """

        def snap_future_callback(pkt):
            source_addr = pkt['src']
            args = pkt['args']

            log.debug("%s callback %s%r received", source_addr, funcname, args)

            log.debug("CALLBACK FUTURES: %r", self._callback_futures)
            log.debug("LOOKING FOR: %s.%s", source_addr, funcname)

            f = self._callback_futures.pop("%s.%s" % (source_addr, funcname), None)

            if not f:
                # Try to see if there's a non-address specific future waiting for multicast
                f = self._callback_futures.get(funcname, None)

            log.debug("FUTURE FOUND: %r", f)

            if f and f.running():
                if f.is_multicast:
                    # See if we already heard this one
                    if source_addr in f.responses:
                        # If we're receiving a duplicate at the application level, make sure to log this behavior
                        log.error("Received duplicate mcast for %s from %s", funcname, source_addr)
                    elif source_addr not in f.expected_responses:
                        log.error("Received unexpected mcast for %s from %s", funcname, source_addr)
                    else:
                        # If we have them all, resolve the future
                        f.responses[source_addr] = args
                        f.expected_responses.remove(source_addr)

                        if not f.expected_responses:
                            del self._callback_futures[funcname]
                            self.unsubscribe_all(funcname)
                            f.set_result(f.responses)
                else:
                    self.unsubscribe_all(funcname)
                    f.set_result(args)

            else:
                log.warning("Received callback from %s for %s%r but no future available", source_addr, funcname, args)

        return snap_future_callback

    def _register_callback(self, funcname):
        """Register an RPC callback that can resolve futures

        :param str funcname: Name of function to resolve
        """
        log.info("Registering callback for func: %s", funcname)
        self.subscribe_rpc(funcname, self._get_callback(funcname))

    @coroutine
    def _rpc_with_retries(self, message_type, funcname, args, addrs, group, ttl, dmcast_delay):
        """Return a future that attempts to send any type of RPC multiple times

        If the RPC could never be sent out the system, the Future resolves to False.
        If the RPC could have been sent, the Future resolves to True.

        Note: These retries are not the same as trying to get a response! This is just attempting to send the message.

        :param str message_type: MESSAGE_TYPE_MCAST, MESSAGE_TYPE_DMCAST, or MESSAGE_TYPE_RPC
        :param list args: Arguments passed to the RPC
        :param str|list[str] addrs: Address or addresses to send the message to if applicable
        :param int|None ttl: TTL if applicable
        :param int|None dmcast_delay: DMCast delay if applicable
        """
        for rpc_retry in xrange(RPC_RETRIES + 1):
            if message_type == MESSAGE_TYPE_MCAST:
                rpc_success = yield self.mRPC(group, ttl, *args)
            elif message_type == MESSAGE_TYPE_RPC:
                rpc_success = yield self.uRPC(addrs, *args)
            elif message_type == MESSAGE_TYPE_DMCAST:
                rpc_success = yield self.dmRPC(addrs, group, ttl, dmcast_delay, *args)

            if not rpc_success:
                log.error("RPC failed to send for %s %s%r", addrs if message_type == MESSAGE_TYPE_RPC else message_type, funcname, args)
            else:
                raise Return(True)

        raise Return(False)

    @staticmethod
    def get_results_with_result_type(result_type, callback_result):
        """Return the responses that match a result type.

        :param namedtuple result_type: The type of callback expected
        :param dict[str,tuple]|tuple callback_result: The result of the callback

        :rtype: tuple(dict, dict)
        :returns: The converted results and the dictionary of invalid responses that were too long or too short
        """
        invalid_responses = {}
        log.debug("Result Type: %r", result_type)
        log.debug("Callback Result: %r", callback_result)

        if result_type:
            # If a namedtuple was provided, convert tuple results to the named variety
            if isinstance(callback_result, dict):
                # Responses with too many or too few arguments when a callback_result is provided are treated as failures
                invalid_responses = {k: v for k, v in callback_result.iteritems() if len(v) != len(result_type._fields)}
                callback_result = {k: result_type(*v) if k not in invalid_responses else None for k, v in callback_result.iteritems()}
            else:
                callback_result = result_type(*callback_result) if len(callback_result) == len(result_type._fields) else None

        log.debug("Result, Responses = %r, %r", callback_result, invalid_responses)

        return callback_result, invalid_responses

    @coroutine  # noqa
    def _callback_message(self, addrs, funcname, args=(), timeout=2, retries=3,
                          result_type=None, callback_name=None, message_type=None,
                          group=DEFAULT_MCAST_GROUP, ttl=DEFAULT_MCAST_TTL,
                          dmcast_delay=DEFAULT_DMCAST_DELAY, node_limit=None):
        """Returns a future for the result, or None if there was a timeout

        Sends mcast, dmcast, and RPC messages

        :param str|list[str] addrs: SNAP Address (Ex: "aabbcc") or list of SNAP addresses to expect responses from
        :param str funcname: Name of function to call with RPC
        :param tuple args: Tuple of arguments to pass to the node
        :param float timeout: Number of seconds to wait before timing out
        :param int retries: Number of times to retry the message after a timeout
        :param namedtuple result_type: Named tuple describing the format of the returned value
        :param str|None callback_name: If None, send the RPC with 'callback' to a random callback
                                       If set to a value, expect your RPC to generate a response to
                                       a certain function and do NOT use 'callback' in the RPC
        :param int|None ttl: TTL for multicast packets or None to use the default
        :param int dmcast_delay: milliseconds to delay each response from the targets (dmcast only)
        :param int|None node_limit: The max number of nodes allowed in a dmcast or None if we should fit as many as possible. (dmcast only)

        :returns: A future that resolves to the return value of the function (RPC) a dictionary of addresses and return values (mcast_rpc) or a tuple containing
                  the dictionary of addresses and return values along with a list of unprocessed nodes
        """
        # True if we should use RPC the 'callback' function
        # False when we just expect the function to respond on its own
        use_callback_rpc = False

        # The future would resolve to the response RPCs arguments
        if callback_name is None:
            use_callback_rpc = True
            callback_name = self._get_unique_name(funcname)

        self._register_callback(callback_name)

        if message_type == MESSAGE_TYPE_DMCAST:
            # figure out how many addresses we can put in the packet.
            addrs, unprocessed_nodes = self._get_addrs_within_overhead(addrs, args, funcname, node_limit)

        future = SnapFuture(message_type in [MESSAGE_TYPE_MCAST, MESSAGE_TYPE_DMCAST], addrs, callback_name)
        future_key = future.get_key()
        self._callback_futures[future_key] = future

        log.debug("CALLBACK FUTURES: %r", self._callback_futures)

        if use_callback_rpc:
            # We need 'callback' to get us back to our random function
            callback_args = ('callback', callback_name, funcname) + args
        else:
            # If we have a specific callback, just make the RPC and the function should reply on its own
            callback_args = (funcname,) + args

        # Send an RPC and wait for it to leave the system
        while retries >= 0:
            # Log the callback information. Ex: "Callback RPC dmcast.some_function(arg1, arg2, arrg3) to ['001122'] timeout=3 retries=0"
            message_destination = addrs if message_type == MESSAGE_TYPE_RPC else message_type  # ex: 'aa0011', 'dmcast', 'mcast'
            log_responses = " to %s" % future.expected_responses if message_type == MESSAGE_TYPE_DMCAST else ""  # Only show expected responses for dmcast
            log.info("Callback RPC %s.%s%r%s timeout=%s retries=%s", message_destination, funcname, args, log_responses, timeout, retries)

            # RPC retries are handled separately from the response retries
            rpc_send_result = yield self._rpc_with_retries(message_type, funcname, callback_args, future.expected_responses, group, ttl, dmcast_delay)

            if not rpc_send_result:
                # If we couldn't even send the message, there's no point in waiting for any replies.
                # Resolve the future with whatever we have outside of the loop
                log.error("Unable to send RPC after multiple attempts. Giving up and returning responses")
                break

            try:
                callback_result = yield with_timeout(timedelta(seconds=timeout), future)
            except TimeoutError:
                log.error("RPC timed out for %s %s%r (Retries=%d)", addrs if message_type == MESSAGE_TYPE_RPC else message_type, funcname, args, retries)
            else:
                # If we didn't have a timeout error, break out of the loop
                # Map from address to the invalid response tuple. In the responses dict, these will be set to None as if they did not respond.
                callback_result, invalid_responses = self.get_results_with_result_type(result_type, callback_result)

                if invalid_responses:
                    log.error("RPC received invalid responses for %s %s%r: %r",
                              addrs if message_type == MESSAGE_TYPE_RPC else message_type,
                              funcname, args, invalid_responses)

                if message_type == MESSAGE_TYPE_DMCAST:
                    # dmcast messages further wrap the responses to include unprocessed nodes
                    callback_result = MultiResult(callback_result, unprocessed_nodes, invalid_responses)

                raise Return(callback_result)

            retries -= 1

        # If we reached this point, we're out of retries, get rid of this future, we don't need it
        self._callback_futures.pop(future_key, None)
        self.unsubscribe_all(callback_name)

        if isinstance(addrs, list):
            # If we're multicast or dmcast, we can still return if we heard from anyone, even if we timed out and never heard from everybody
            future_responses, invalid_responses = self.get_results_with_result_type(result_type, future.responses)
            response = {a: None for a in addrs}
            response.update(future_responses)

            if invalid_responses:
                log.error("RPC received invalid responses for %s %s%r: %r",
                          addrs if message_type == MESSAGE_TYPE_RPC else message_type,
                          funcname, args, invalid_responses)

            if message_type == MESSAGE_TYPE_DMCAST:
                response = MultiResult(response, unprocessed_nodes, invalid_responses)

            raise Return(response)
        else:
            # If we weren't a multicast or dmcast message, the future resolves to None at this point to indicate no response
            pass

    @staticmethod
    def _get_addrs_within_overhead(addrs, args, funcname, node_limit=None):
        """Return the address that will fit into a given dmcast packet and those that will be unprocessed.

        :param list[str] addrs: List of addresses (ex: ['aabbcc'])
        :param list args: Arguments passed to the function
        :param str funcname: Function being called
        :param int node_limit: Number of nodes to use as a maximum or None if we want to use as many as possible
        """
        # Find out how many addresses will fit
        max_addrs = get_max_addrs(get_dmcast_packet_overhead(funcname, '', *args))

        # Now, take the smaller of the node_limit and the addrs that will actually fit
        if node_limit is not None:
            max_addrs = min(max_addrs, node_limit)
        if max_addrs < len(addrs):
            log_rpc.debug("Only %d of %d addrs will fit in the next DMcast packet", max_addrs, len(addrs))

        unprocessed_nodes = addrs[max_addrs:]  # List of nodes that wouldn't fit in this message
        processed_nodes = addrs[:max_addrs]  # List of nodes that will fit into this message

        return processed_nodes, unprocessed_nodes

    def callback_rpc(self, addr, funcname, args=(), timeout=2, retries=3, result_type=None, callback_name=None):
        """Returns a future for the result, or None if there was a timeout

        :param str addr: SNAP Address (Ex: "aabbcc")
        :param str funcname: Name of function to call with RPC
        :param tuple args: Tuple of arguments to pass to the node
        :param float timeout: Number of seconds to wait before timing out
        :param int retries: Number of times to retry the message after a timeout
        :param namedtuple result_type: Named tuple describing the format of the returned value
        :param str|None callback_name: If None, send the RPC with 'callback' to a random callback
                                       If set to a value, expect your RPC to generate a response to
                                       a certain function and do NOT use 'callback' in the RPC
        :rtype: Future
        :returns: A future that resolves to the return value of the function or None if we don't get a result
        """
        return self._callback_message(addr, funcname, args, timeout=timeout, retries=retries, result_type=result_type,
                                      callback_name=callback_name, message_type=MESSAGE_TYPE_RPC)

    def callback_mcast_rpc(self, addrs, funcname, args=(), timeout=2, retries=3, result_type=None, callback_name=None, group=None, ttl=None):
        """Returns a future for the results, or None if there was a timeout

        This waits for every address to respond, or until we're out of timeouts and retries, then we give up

        :param str|list(str) addrs: List of SNAP Address we expect to respond (Ex: ["aabbcc"])
        :param str funcname: Name of function to call with RPC
        :param tuple args: Tuple of arguments to pass to the node
        :param float timeout: Number of seconds to wait before timing out
        :param int retries: Number of times to retry the message after a timeout
        :param namedtuple result_type: Named tuple describing the format of the returned value
        :param str|None callback_name: If None, send the RPC with 'callback' to a random callback
                                       If set to a value, expect your RPC to generate a response to
                                       a certain function and do NOT use 'callback' in the RPC
        :param int|None ttl: TTL for multicast packets or None to use the default

        :rtype: Future
        :returns: A future that resolves to the return value of the functions as a dictionary or None if we don't get a result
        """
        if isinstance(addrs, basestring):
            addrs = [addrs]

        return self._callback_message(addrs, funcname, args, timeout=timeout, retries=retries, result_type=result_type,
                                      callback_name=callback_name, group=group, ttl=ttl, message_type=MESSAGE_TYPE_MCAST)

    @staticmethod
    def _sort_by_attempts_remaining(attempts_remaining):
        """Return a sorted list of addresses with the most attempts remaining first

        :param dict[str, int] attempts_remaining: Dict from address to attempts remaining
        :rtype: list
        """
        sorted_list = attempts_remaining.items()
        sorted_list.sort(key=lambda tup: tup[1], reverse=True)
        return [tup[0] for tup in sorted_list]

    @coroutine
    def callback_dmcast(self, addrs, funcname, args=(), timeout=2, retries=3,
                        result_type=None, callback_name=None, dmcast_delay=DEFAULT_DMCAST_DELAY,
                        node_limit=None, group=DEFAULT_MCAST_GROUP, ttl=DEFAULT_MCAST_TTL,
                        chunk_results=False, previous_chunk=None):
        """Returns a future for the results, or None if there was a timeout

        This waits for every address to respond, or until we're out of timeouts and retries, then we give up

        Example response:
        (
            {
                "aabbcc": ("response1,), # This one responded with one argument
                "bbccdd": (10, 20 30), # This one responded with 3 arguments
                "ddeeff": None,  # We didn't hear a response from this one
            },
            ["aabb11", "bbcc22"]  # We couldn't fit these 2
        )

        :param str|list(str) addrs: List of SNAP Address we expect to respond (Ex: ["aabbcc"])
        :param str funcname: Name of function to call with RPC
        :param tuple args: Tuple of arguments to pass to the node
        :param float timeout: Number of seconds to wait before timing out
        :param int retries: Number of times to retry each address after a timeout. This is NOT the number of messages.
        :param namedtuple result_type: Named tuple describing the format of the returned value
        :param str|None callback_name: If None, send the RPC with 'callback' to a random callback
                                       If set to a value, expect your RPC to generate a response to
                                       a certain function and do NOT use 'callback' in the RPC
        :param int dmcast_delay: milliseconds to delay each response from the targets
        :param int|None node_limit: The max number of nodes allowed in a dmcast or None if we should fit as many as possible.
        :param bool chunk_results: If True, return the results in chunks with unprocessed nodes.
                                   If False, run until we have every single result and then return.
        :param MultiResult previous_chunk: Used to continue processing data when chunk_results is set to True.
                                           Used internally when `chunk_results` is set to True.
        :rtype: Future
        :returns: A future that resolves to the return value of the functions as a dictionary None if we don't get a result
        """
        # NOTE: This uses the 2.6 builtin dmcast_rpc - you must be using correct snapconnect / snaplib for this to work.
        if isinstance(addrs, basestring):
            addrs = [addrs]

        if previous_chunk is None:
            # Map from address to attempts remaining. OrderedDict preserves passed in order.
            # Set the initial number of attempts allowed for each node
            attempts_remaining = OrderedDict([(addr, retries+1) for addr in addrs])
            current_results = {}  # Map from addr to the response object
            current_invalid_results = {}
        else:
            attempts_remaining = previous_chunk.attempts_remaining
            current_results = previous_chunk.responses
            current_invalid_results = previous_chunk.invalid_responses

        multi_response = None

        while attempts_remaining:
            addrs = self._sort_by_attempts_remaining(attempts_remaining)

            # Because we're managing retries in this method, set them to 0 in _callback_message
            multi_response = yield self._callback_message(addrs, funcname, args, timeout=timeout, retries=0, result_type=result_type,
                                                          callback_name=callback_name, group=group, ttl=ttl, message_type=MESSAGE_TYPE_DMCAST,
                                                          dmcast_delay=dmcast_delay, node_limit=node_limit)

            current_results.update(multi_response.responses)
            current_invalid_results.update(multi_response.invalid_responses)

            # Remove any successful nodes from attempts_remaining
            for addr, result in multi_response.responses.iteritems():
                if result is None and addr not in multi_response.invalid_responses:
                    # This is a failed attempt
                    attempts_remaining[addr] -= 1

                    if attempts_remaining[addr] == 0:
                        # Remove this from the list, we're done with it
                        attempts_remaining.pop(addr, 0)
                else:
                    # Remove this from the list, we heard from the node or it was invalid and we do not need to retry it
                    attempts_remaining.pop(addr, 0)

            if chunk_results:
                # If we're returning in chunks, we don't want to continue in this loop
                # Instead, we attach our state to MutliResult object and allow it to call a special .retry() function
                multi_response.chunk_responses = multi_response.responses.copy()  # Most recent responses are the chunk responses
                multi_response.attempts_remaining = attempts_remaining

                def retry_function():
                    return self.callback_dmcast(addrs, funcname, args, timeout=timeout, retries=0, result_type=result_type, callback_name=callback_name,
                                                dmcast_delay=dmcast_delay, node_limit=node_limit, group=group, ttl=ttl,
                                                chunk_results=chunk_results, previous_chunk=multi_response)

                multi_response.set_retry(retry_function)
                break  # Do not continue looping when chunking the results

        if multi_response:
            # Roll our most recent responses and failures into the totals.
            multi_response.responses.update(current_results)
            multi_response.invalid_responses.update(current_invalid_results)
            raise Return(multi_response)
