"""Utilities for testing future Snap2MQTT"""
import logging
from itertools import cycle

import mock
from tornado.testing import AsyncTestCase
from tornado.gen import coroutine, moment
from tornado.concurrent import Future

import paho.mqtt.client as mqtt
import msgpack

from snap2mqtt_client.client import BaseClient

import snap2mqtt_future_client
from future_client import FutureClient


log = logging.getLogger(__name__)


class FutureClientTestCase(AsyncTestCase):
    """Test that the BaseClient futures wrapper works correctly"""

    def setUp(self):
        super(FutureClientTestCase, self).setUp()
        self.fsc = self.setUpFsc()
        self.client = self.fsc._client
        self.local_addr = self.client.router_address

        assert isinstance(self.fsc, FutureClient)

    def setUpFsc(self):
        """Setup the FSC instance for the test

        This may be overriden by subclasses to alter parts of the setup if desired.

        This function MUST return a valid fsc object
        """
        client = get_mock_client()
        fsc = FutureClient(client)
        fsc._app_seq = 0

        def mock_tx(mid, success=True):
            if success:
                fsc._on_publish(None, None, mid)
            else:
                try:
                    fsc._rpc_futures[mid].set_result(False)
                    del fsc._rpc_futures[mid]
                    fsc._message_sent()
                except KeyError:
                    snap2mqtt_future_client.future_client.log_rpc.info("Could not find an existing Future for Message id %s", mid)
                    snap2mqtt_future_client.future_client.log_rpc.debug("Outstanding RPC Futures: %r", fsc._rpc_futures.keys())

        fsc.mock_tx = mock_tx
        client.mock_rpc_sent = mock_tx

        return fsc

    def assertRpcCalls(self, mock_calls):
        """Asert that a given list of mock calls was made to RPC"""
        self.assertEqual(self.client.send.mock_calls, mock_calls)
        self.client.send.reset_mock()

    def assertMcastCalls(self, mock_calls):
        """Asert that a given list of mock calls was made via mcast"""
        self.assertEqual(self.client.send.mock_calls, mock_calls)
        self.client.send.reset_mock()

    def assertDMcastCalls(self, mock_calls):
        """Asert that a given list of mock calls was made via dmcast"""
        self.assertEqual(self.client.send.mock_calls, mock_calls)
        self.client.send.reset_mock()


@coroutine
def moments():
    """Burn through all coroutines on the ioloop (Not timeouts)"""
    # TODO - Is there something in the ioloop we can look at to know the real number?
    for x in xrange(10):
        yield moment


def get_mock_client():
    """Return a mock BaseClient object"""
    client = BaseClient('000001')

    def mock_rx(addr, name, *args):
        """Mock the receipt of an RPC from a node

        :param addr: Hexlified SNAP Address (Ex: 'aabbcc')
        :param name: Name of RPC to call
        :param *args: Arguments for the RPC
        """

        msg = mqtt.MQTTMessage()
        msg.topic = 'snap/000001/recv/rpc/' + addr + '/' + name
        msg.payload = msgpack.dumps({'type': 'uRPC', 'src': addr, 'func': name, 'args': args})
        client._on_mqtt_message(None, None, msg)

    client.mock_rx = mock_rx

    def mock_rpc_sent(mid, success=True):
        """Mock the transmission of an RPC to a node

        :param mid: Packet message id
        :param success (bool|None): Whether the packet transmitted successfully
        """

        client._on_mqtt_publish(mid)

    client.mock_rpc_sent = mock_rpc_sent

    packet_sequence = cycle(xrange(255))

    def packet_id():
        future = Future()
        future.set_result(next(packet_sequence))
        return

    client.callback_rpc = mock.Mock()
    client.callback_rpc.side_effect = packet_id
    client.callback_dmcast = mock.Mock()
    client.callback_dmcast.side_effect = packet_id
    client.callback_mcast = mock.Mock()
    client.callback_mcast.side_effect = packet_id
    client.rpc_source_addr = mock.Mock()
    client.rpc_source_addr.return_value = '\xaa\xbb\xcc'

    client.start = mock.Mock()
    client.stop = mock.Mock()
    client.loop = mock.Mock()

    def subscribe():
        pass

    def unsubscribe():
        pass

    def send():
        pass

    mid_seq = cycle(xrange(255))

    client.send = mock.Mock()
    client.send.side_effect = lambda *args: (0, mid_seq.next())

    client._on_mqtt_publish = mock.Mock()

    return client
