"""Test SNAP Connect futures resolve correctly"""
import gc
import traceback
from collections import namedtuple
import logging

import mock
from tornado.gen import sleep, Future
from tornado.testing import gen_test

from .future_client import FutureClient
from .future_client import DEFAULT_DMCAST_DELAY, DEFAULT_MCAST_GROUP, DEFAULT_MCAST_TTL
from .packet_overhead import get_packet_overhead, get_max_addrs
from .responses import MultiResult

from s2mfc_test_case import FutureClientTestCase, get_mock_client, moments


class TestFutureClient(FutureClientTestCase):
    """Test that the futures wrapper works correctly"""

    def tearDown(self):
        # Check for any outstanding futures
        uncollected_futures = [obj for obj in gc.get_objects() if isinstance(obj, Future)]

        for future in uncollected_futures:

            if future._log_traceback:  # pragma: no cover
                # set_exception() was called, or result() or exception() were not called

                tb = traceback.format_exception(*future._exc_info)
                print tb
                self.fail('Future exception for %s was never retrieved: %s' % (future, ''.join(tb).rstrip()))

        super(TestFutureClient, self).tearDown()

    def test_initialization(self):
        """Test that the future class initializes correctly"""

        # Make sure our futures all start in a default state
        self.assertEqual(self.fsc._rpc_futures, {})
        self.assertEqual(self.fsc._callback_futures, {})

    def test_get_unique_name(self):
        """Test that future SNAP Connect can generate unique names for function callbacks"""
        self.assertEqual(self.fsc._get_unique_name("test"), "test1")
        self.assertEqual(self.fsc._get_unique_name("test"), "test2")
        self.assertEqual(self.fsc._get_unique_name("test"), "test3")

        self.assertEqual(self.fsc._get_unique_name("other_test"), "other_test1")
        self.assertEqual(self.fsc._get_unique_name("other_test"), "other_test2")
        self.assertEqual(self.fsc._get_unique_name("other_test"), "other_test3")

    @gen_test
    def test_rpc(self):
        """Test that we can asynchronously send RPCs"""
        # Test a successful RPC
        send_rpc_future = self.fsc.uRPC("aabbcc", "test_func", "arg1", "arg2", "arg3")
        self.assertTrue(send_rpc_future.running())
        self.assertRpcCalls([mock.call({'dst': 'aabbcc', 'args': ('arg1', 'arg2', 'arg3'), 'type': 'uRPC', 'func': 'test_func'})])

        # Make sure we save the future off
        self.assertEqual(self.fsc._rpc_futures, {0: send_rpc_future})

        # Simulate a successful RPC exiting the system
        self.client.mock_rpc_sent(0)
        self.assertEqual(self.fsc._rpc_futures, {})

        # At this point, our future has resolved, so we have a result
        rpc_result = yield send_rpc_future
        self.assertEqual(rpc_result, True)

        # Try again, where the RPC fails
        send_rpc_future = self.fsc.uRPC("aabbcc", "test_func")
        self.assertTrue(send_rpc_future.running())
        self.assertRpcCalls([mock.call({'dst': 'aabbcc', 'args': (), 'type': 'uRPC', 'func': 'test_func'})])

        # Make sure we save the future off
        self.assertEqual(self.fsc._rpc_futures, {1: send_rpc_future})

        # Simulate an RPC exiting the system failure
        self.client.mock_rpc_sent(1, success=False)  # Packet ID 1 returned False
        self.assertEqual(self.fsc._rpc_futures, {})

        # At this point, our future has resolved, so we have a result
        rpc_result = yield send_rpc_future
        self.assertEqual(rpc_result, False)

    @gen_test
    def test_multiple_rpc(self):
        """Test that we can asynchronously send multiple RPCs in different orders"""
        # Try with multiple simultaneously
        send_rpc_future_1 = self.fsc.uRPC("aabbcc", "test_func", "arg1", "arg2", "arg3")
        self.assertTrue(send_rpc_future_1.running())
        self.assertRpcCalls([mock.call({'dst': 'aabbcc', 'args': ('arg1', 'arg2', 'arg3'), 'type': 'uRPC', 'func': 'test_func'})])

        send_rpc_future_2 = self.fsc.uRPC("aabbcc", "test_func", "arg1_2", "arg2_2", "arg3_2")
        self.assertTrue(send_rpc_future_2.running())
        self.assertRpcCalls([mock.call({'dst': 'aabbcc', 'args': ('arg1_2', 'arg2_2', 'arg3_2'), 'type': 'uRPC', 'func': 'test_func'})])

        # Make sure we save the future off
        self.assertEqual(self.fsc._rpc_futures, {0: send_rpc_future_1, 1: send_rpc_future_2})

        # Simulate a successful RPC exiting the system
        self.client.mock_rpc_sent(1)  # Packet ID 3 returned True
        self.assertEqual(self.fsc._rpc_futures, {0: send_rpc_future_1})

        rpc_result = yield send_rpc_future_2
        self.assertEqual(rpc_result, True)

        # Simulate the other successful RPC exiting the system
        self.client.mock_rpc_sent(0)  # Packet ID 2 returned True
        self.assertEqual(self.fsc._rpc_futures, {})

        rpc_result = yield send_rpc_future_1
        self.assertEqual(rpc_result, True)

    def test_invalid_rpc_received(self):
        """Test that receiving a non existing RPC fails gracefully."""
        with mock.patch("snap2mqtt_future_client.future_client.log_rpc") as mock_log:
            self.client.mock_rpc_sent(6)
            self.assertEqual(mock_log.mock_calls, [
                mock.call.info("Could not find an existing Future for Message id %s", 6), mock.call.debug('Outstanding RPC futures: %r', [])
            ])

    @gen_test
    def test_callback_rpc(self):
        """Test that callback RPC correctly resolves futures when we receive a callback"""
        callback_rpc_future = self.fsc.callback_rpc("aabbcc", "test_func", ("a", "b", "c"))
        self.assertRpcCalls([mock.call({'dst': 'aabbcc', 'args': ('test_func1', 'test_func', 'a', 'b', 'c'), 'type': 'uRPC', 'func': 'callback'})])

        # Make sure we're waiting on this particular RPC
        self.assertEqual(self.fsc._callback_futures.keys(), ["aabbcc.test_func1"])

        # Mock out the RPC being sent
        self.client.mock_rpc_sent(0)

        # Mock out the response to the RPC
        self.client.mock_rx("aabbcc", "test_func1", "Result")

        # Make sure we immediately resolve the future
        yield moments()
        self.assertEqual(callback_rpc_future.running(), False)

        rpc_result = yield callback_rpc_future
        self.assertEqual(rpc_result, ["Result"])

        # Make sure we removed the function from the dispatcher
        self.assertNotIn("rpc/+/test_func1", self.client.subscriptions)

        # Make sure we're not waiting on anything else
        self.assertEqual(self.fsc._callback_futures, {})
        self.assertEqual(self.fsc._rpc_futures, {})
        self.assertRpcCalls([])

    @gen_test
    def test_callback_rpc_invalid_response(self):
        """Test that the wrong address does not resolve the future"""
        callback_rpc_future = self.fsc.callback_rpc("aabbcc", "test_func", ("a", "b", "c"))

        # Mock out the RPC being sent and responded to
        self.client.mock_rpc_sent(0)
        self.client.mock_rx("badadd", "test_func1", "Result")

        # Make sure we didn't immediately resolve the future for either case
        yield moments()
        self.assertEqual(callback_rpc_future.running(), True)

        # Make sure we did not remove the function from the dispatcher
        self.assertIn("rpc/+/test_func1", self.client.subscriptions)

    @gen_test
    def test_callback_rpc_no_arguments(self):
        """Test that callback RPC correctly resolves futures when we receive a callback with no arguments"""
        callback_rpc_future = self.fsc.callback_rpc("aabbcc", "test_func")

        # Mock out the RPC being sent and responded to
        self.client.mock_rpc_sent(0)
        self.client.mock_rx("aabbcc", "test_func1")

        rpc_result = yield callback_rpc_future
        self.assertEqual(rpc_result, [])

    @gen_test
    def test_callback_rpc_with_result_tuple(self):
        """Test that callback RPC correctly resolves futures with a result tuple when we receive a callback"""
        example_type = namedtuple("ExampleType", ["a", "b", "c"])
        callback_rpc_future = self.fsc.callback_rpc("aabbcc", "test_func", result_type=example_type)

        # Mock out the RPC being sent and responded to
        self.client.mock_rpc_sent(0)
        self.client.mock_rx("aabbcc", "test_func1", "Result1", "Result2", "Result3")

        rpc_result = yield callback_rpc_future
        self.assertEqual(rpc_result, example_type("Result1", "Result2", "Result3"))
        self.assertIsInstance(rpc_result, example_type)

    @gen_test
    def test_callback_rpc_timeout(self):
        """Test that callback RPC correctly resolves futures when we receive a timeout"""
        callback_rpc_future = self.fsc.callback_rpc("aabbcc", "test_func", ("a", "b", "c"), retries=2, timeout=.1)
        self.assertRpcCalls([mock.call({'dst': 'aabbcc', 'args': ('test_func1', 'test_func', 'a', 'b', 'c'), 'type': 'uRPC', 'func': 'callback'})])
        self.client.mock_rpc_sent(0)

        # After waiting half the time we shouldn't have done anything
        yield sleep(.05)
        self.assertRpcCalls([])

        # Make sure we retried (1x)
        yield sleep(.06)
        self.assertRpcCalls([mock.call({'dst': 'aabbcc', 'args': ('test_func1', 'test_func', 'a', 'b', 'c'), 'type': 'uRPC', 'func': 'callback'})])
        self.client.mock_rpc_sent(1)

        # After waiting half the time we shouldn't have done anything
        yield sleep(.05)
        self.assertRpcCalls([])

        # Make sure we retried (2x)
        yield sleep(.06)
        self.assertRpcCalls([mock.call({'dst': 'aabbcc', 'args': ('test_func1', 'test_func', 'a', 'b', 'c'), 'type': 'uRPC', 'func': 'callback'})])
        self.client.mock_rpc_sent(2)

        # After waiting half the time we shouldn't have done anything
        yield sleep(.05)
        self.assertRpcCalls([])

        # Make sure after the final timeout we return None
        yield sleep(.06)
        self.assertEqual(callback_rpc_future.running(), False)

        rpc_result = yield callback_rpc_future
        self.assertEqual(rpc_result, None)

        # Make sure we removed the function from the dispatcher
        self.assertNotIn("rpc/+/test_func1", self.client.subscriptions)

        # Make sure we're not waiting on anything else
        self.assertEqual(self.fsc._callback_futures, {})
        self.assertEqual(self.fsc._rpc_futures, {})
        self.assertRpcCalls([])

    @gen_test
    def test_callback_rpc_retries(self):
        """Test that callback RPC correctly resolves futures when we have retries"""
        callback_rpc_future = self.fsc.callback_rpc("aabbcc", "test_func", ("a", "b", "c"), retries=1, timeout=.05)
        self.assertRpcCalls([mock.call({'dst': 'aabbcc', 'args': ('test_func1', 'test_func', 'a', 'b', 'c'), 'type': 'uRPC', 'func': 'callback'})])
        self.client.mock_rpc_sent(0)

        # Wait .05+ without receiving the response!
        yield sleep(.06)

        # Make sure we retried
        self.assertRpcCalls([mock.call({'dst': 'aabbcc', 'args': ('test_func1', 'test_func', 'a', 'b', 'c'), 'type': 'uRPC', 'func': 'callback'})])
        self.client.mock_rpc_sent(1)

        # Mock out the response to the RPC
        self.client.mock_rx("aabbcc", "test_func1", "Result")

        # Make sure we immediately resolve the future
        yield moments()
        self.assertEqual(callback_rpc_future.running(), False)

        rpc_result = yield callback_rpc_future
        self.assertEqual(rpc_result, ["Result"])

        # Make sure we're not waiting on anything else
        self.assertEqual(self.fsc._callback_futures, {})
        self.assertEqual(self.fsc._rpc_futures, {})
        self.assertRpcCalls([])

    @gen_test
    def test_multiple_callback_rpc(self):
        """Test that can have multiple callback RPC calls outstanding"""
        callback_rpc_future_1 = self.fsc.callback_rpc("aabbcc", "test_func", ("a", "b", "c"))
        self.assertRpcCalls([mock.call({'dst': 'aabbcc', 'args': ('test_func1', 'test_func', 'a', 'b', 'c'), 'type': 'uRPC', 'func': 'callback'})])
        self.client.mock_rpc_sent(0)

        callback_rpc_future_2 = self.fsc.callback_rpc("ffaabb", "test_func", ("c", "d", "e"))
        self.assertRpcCalls([mock.call({'dst': 'ffaabb', 'args': ('test_func2', 'test_func', 'c', 'd', 'e'), 'type': 'uRPC', 'func': 'callback'})])
        self.client.mock_rpc_sent(1)

        logging.info(self.fsc._callback_futures)

        # Mock out responses in the opposite order
        self.client.mock_rx("ffaabb", "test_func2", "Result ffaabb")
        self.client.mock_rx("aabbcc", "test_func1", "Result aabbcc")

        rpc_result = yield callback_rpc_future_1
        self.assertEqual(rpc_result, ["Result aabbcc"])

        rpc_result = yield callback_rpc_future_2
        self.assertEqual(rpc_result, ["Result ffaabb"])

        # Make sure we removed the function from the dispatcher
        self.assertNotIn("rpc/+/test_func1", self.client.subscriptions)

    @gen_test
    def test_callback_rpc_with_failed_rpc(self):
        """Test callback RPC where the RPC fails to send"""

        # Fake send failures
        self.fsc._client.send.side_effect = lambda *args: (1, 0)

        callback_rpc_future = self.fsc.callback_rpc("aabbcc", "test_func", ("a", "b", "c"), timeout=.1)
        self.assertRpcCalls([mock.call({'dst': 'aabbcc',
                                        'args': ('test_func1', 'test_func', 'a', 'b', 'c'),
                                        'type': 'uRPC',
                                        'func': 'callback'}),
                             mock.call({'dst': 'aabbcc',
                                        'args': ('test_func1', 'test_func', 'a', 'b', 'c'),
                                        'type': 'uRPC',
                                        'func': 'callback'}), mock.call(
                                            {'dst': 'aabbcc',
                                             'args': ('test_func1', 'test_func', 'a', 'b', 'c'),
                                             'type': 'uRPC',
                                             'func': 'callback'}), mock.call(
                                                 {'dst': 'aabbcc',
                                                  'args': ('test_func1', 'test_func', 'a', 'b', 'c'),
                                                  'type': 'uRPC',
                                                  'func': 'callback'})])

        # There is no callback on failure to send in MQTT, you just have to time out

        # Make sure after the final timeout we return None
        yield moments()
        self.assertEqual(callback_rpc_future.running(), False)

        rpc_result = yield callback_rpc_future
        self.assertEqual(rpc_result, None)

        # Make sure we're not waiting on anything else
        self.assertEqual(self.fsc._callback_futures, {})
        self.assertEqual(self.fsc._rpc_futures, {})
        self.assertRpcCalls([])

        # Make sure we removed the function from the dispatcher
        self.assertNotIn("rpc/+/test_func1", self.client.subscriptions)

    @gen_test
    def test_callback_dmcast_with_failed_rpc(self):
        """Test callback dmcast where the RPC fails to send"""

        # Fake send failures
        self.fsc._client.send.side_effect = lambda *args: (1, 0)

        callback_rpc_future = self.fsc.callback_dmcast("aabbcc", "test_func", ("a", "b", "c"), retries=0)
        self.assertDMcastCalls([
            mock.call({
                'group': 1,
                'dst': ['aabbcc'],
                'args': ('test_func1', 'test_func', 'a', 'b', 'c'),
                'delay': 50,
                'func': 'callback',
                'ttl': 5,
                'type': 'dmRPC'
            }),
            mock.call({
                'group': 1,
                'dst': ['aabbcc'],
                'args': ('test_func1', 'test_func', 'a', 'b', 'c'),
                'delay': 50,
                'func': 'callback',
                'ttl': 5,
                'type': 'dmRPC'
            }),
            mock.call({
                'group': 1,
                'dst': ['aabbcc'],
                'args': ('test_func1', 'test_func', 'a', 'b', 'c'),
                'delay': 50,
                'func': 'callback',
                'ttl': 5,
                'type': 'dmRPC'
            }),
            mock.call({
                'group': 1,
                'dst': ['aabbcc'],
                'args': ('test_func1', 'test_func', 'a', 'b', 'c'),
                'delay': 50,
                'func': 'callback',
                'ttl': 5,
                'type': 'dmRPC'
            }),
        ])

        # There is no callback on failure to send in MQTT, you just have to time out

        # Make sure after the final timeout we return None
        yield moments()
        self.assertEqual(callback_rpc_future.running(), False)

        rpc_result = yield callback_rpc_future
        self.assertEqual(rpc_result.responses, {"aabbcc": None})

        # Make sure we're not waiting on anything else
        self.assertEqual(self.fsc._callback_futures, {})
        self.assertEqual(self.fsc._rpc_futures, {})
        self.assertRpcCalls([])

        # Make sure we removed the function from the dispatcher
        self.assertNotIn("rpc/+/test_func1", self.client.subscriptions)

    @gen_test
    def test_callback_rpc_vmstat(self):
        """Test that can have callback RPC expect different RPCs in response (Ex: vmStat)"""
        callback_rpc_future = self.fsc.callback_rpc("aabbcc", "vmStat", (1, 2), callback_name="tellVmStat")
        self.assertRpcCalls([mock.call({'dst': 'aabbcc', 'args': (1, 2), 'type': 'uRPC', 'func': 'vmStat'})])

        # Make sure we're waiting on this particular RPC
        self.assertEqual(self.fsc._callback_futures.keys(), ["aabbcc.tellVmStat"])

        # Mock out the RPC being sent
        self.client.mock_rpc_sent(0)

        # Mock out the response to the RPC
        self.client.mock_rx("aabbcc", "tellVmStat", "Result")

        # Make sure we immediately resolve the future
        yield moments()
        self.assertEqual(callback_rpc_future.running(), False)

        rpc_result = yield callback_rpc_future
        self.assertEqual(rpc_result, ["Result"])

        # Make sure we're not waiting on anything else
        self.assertEqual(self.fsc._callback_futures, {})
        self.assertEqual(self.fsc._rpc_futures, {})
        self.assertRpcCalls([])

        # Make sure we removed the function from the dispatcher
        self.assertNotIn("rpc/+/tellVmStat", self.client.subscriptions)

    @gen_test
    def test_mcast_rpc(self):
        """Test that we can asynchronously send mcast RPCs"""
        # Test a successful RPC
        send_rpc_future = self.fsc.mRPC(DEFAULT_MCAST_GROUP, DEFAULT_MCAST_TTL, "test_func", "arg1", "arg2", "arg3")
        self.assertTrue(send_rpc_future.running())
        self.assertMcastCalls([
            mock.call({'group': DEFAULT_MCAST_GROUP,
                       'args': ('arg1', 'arg2', 'arg3'),
                       'type': 'mRPC',
                       'func': 'test_func',
                       'ttl': DEFAULT_MCAST_TTL})
        ])

        # Make sure we save the future off
        self.assertEqual(self.fsc._rpc_futures, {0: send_rpc_future})

        # Simulate a successful RPC exiting the system
        self.client.mock_rpc_sent(0)  # Packet ID 0 returned True
        self.assertEqual(self.fsc._rpc_futures, {})

        # At this point, our future has resolved, so we have a result
        rpc_result = yield send_rpc_future
        self.assertEqual(rpc_result, True)

        # Try again, where the RPC fails
        send_rpc_future = self.fsc.mRPC(DEFAULT_MCAST_GROUP, DEFAULT_MCAST_TTL, "test_func")
        self.assertTrue(send_rpc_future.running())
        self.assertMcastCalls([mock.call({'group': DEFAULT_MCAST_GROUP, 'args': (), 'type': 'mRPC', 'func': 'test_func', 'ttl': DEFAULT_MCAST_TTL})])

        # Make sure we save the future off
        self.assertEqual(self.fsc._rpc_futures, {1: send_rpc_future})

        # Simulate an RPC exiting the system failure
        self.client.mock_rpc_sent(1, success=False)  # Packet ID 1 returned False
        self.assertEqual(self.fsc._rpc_futures, {})

        # At this point, our future has resolved, so we have a result
        rpc_result = yield send_rpc_future
        self.assertEqual(rpc_result, False)

    @gen_test
    def test_callback_mcast_rpc(self):
        """Test that we can asynchronously send mcast RPCs with a callback"""
        callback_rpc_future = self.fsc.callback_mcast_rpc("aabbcc", "test_func", ("a", "b", "c"))
        self.assertMcastCalls([mock.call({'group': None, 'args': ('test_func1', 'test_func', 'a', 'b', 'c'), 'type': 'mRPC', 'func': 'callback', 'ttl': None})])

        # Make sure we're waiting on this particular RPC
        self.assertEqual(self.fsc._callback_futures.keys(), ["test_func1"])

        # Mock out the RPC being sent
        self.client.mock_rpc_sent(0)

        # Mock out the response to the RPC
        self.client.mock_rx("aabbcc", "test_func1", "Result")

        # Make sure we immediately resolve the future
        yield moments()
        self.assertEqual(callback_rpc_future.running(), False)

        rpc_result = yield callback_rpc_future
        self.assertEqual(rpc_result, {"aabbcc": ["Result", ]})

        # Make sure we're not waiting on anything else
        self.assertEqual(self.fsc._callback_futures, {})
        self.assertEqual(self.fsc._rpc_futures, {})
        self.assertRpcCalls([])

        # Make sure we removed the function from the dispatcher
        self.assertNotIn("rpc/+/test_func1", self.client.subscriptions)

    @gen_test
    def test_callback_mcast_rpc_with_callback_name(self):
        """Test that we can asynchronously send mcast RPCs with a callback and a callback name"""
        callback_rpc_future = self.fsc.callback_mcast_rpc("aabbcc", "test_func", ("a", "b", "c"), callback_name="test_func_ack")
        self.assertMcastCalls([mock.call({'group': None, 'args': ('a', 'b', 'c'), 'type': 'mRPC', 'func': 'test_func', 'ttl': None})])

        # Make sure we're waiting on this particular RPC
        self.assertEqual(self.fsc._callback_futures.keys(), ["test_func_ack"])

        # Mock out the RPC being sent
        self.client.mock_rpc_sent(0)

        # Mock out the response to the RPC
        self.client.mock_rx("aabbcc", "test_func_ack", "Result")

        # Make sure we immediately resolve the future
        yield moments()
        self.assertEqual(callback_rpc_future.running(), False)

        rpc_result = yield callback_rpc_future
        self.assertEqual(rpc_result, {"aabbcc": ["Result"]})

        # Make sure we're not waiting on anything else
        self.assertEqual(self.fsc._callback_futures, {})
        self.assertEqual(self.fsc._rpc_futures, {})
        self.assertRpcCalls([])

        # Make sure we removed the function from the dispatcher
        self.assertNotIn("rpc/+/test_func_ack", self.client.subscriptions)

    @gen_test
    def test_callback_mcast_rpc_with_duplicates_and_unexpected(self):
        """Test that we can asynchronously send mcast RPCs with a callback"""
        self.fsc.immediately_resolve_rpcs()

        callback_rpc_future = self.fsc.callback_mcast_rpc(["aabbcc", "ccddee"], "test_func", ("a", "b", "c"))
        self.assertMcastCalls([mock.call({'group': None, 'args': ('test_func1', 'test_func', 'a', 'b', 'c'), 'type': 'mRPC', 'func': 'callback', 'ttl': None})])

        # Mock out the real response to the RPC
        self.client.mock_rx("aabbcc", "test_func1", "Result 1")

        yield moments()

        # Test with a duplicate response
        with mock.patch("snap2mqtt_future_client.future_client.log") as mock_log:
            self.client.mock_rx("aabbcc", "test_func1", "Result")
            self.assertEqual(mock_log.error.mock_calls, [mock.call("Received duplicate mcast for %s from %s", 'test_func1', 'aabbcc')])

        # Test with an unexpected response
        with mock.patch("snap2mqtt_future_client.future_client.log") as mock_log:
            self.client.mock_rx("ccaabb", "test_func1", "Result")
            self.assertEqual(mock_log.error.mock_calls, [mock.call("Received unexpected mcast for %s from %s", 'test_func1', 'ccaabb')])

        # Mock out the final response to the RPC
        self.client.mock_rx("ccddee", "test_func1", "Result 2")

        # Make sure everything resolved correctly
        rpc_result = yield callback_rpc_future
        self.assertEqual(rpc_result, {"aabbcc": ["Result 1"], "ccddee": ["Result 2"]})

        # Make sure we're not waiting on anything else
        self.assertEqual(self.fsc._callback_futures, {})
        self.assertEqual(self.fsc._rpc_futures, {})
        self.assertRpcCalls([])

    @gen_test
    def test_callback_mcast_rpc_with_ttl(self):
        """Test that we can asynchronously send mcast RPCs with a non default ttl"""
        callback_rpc_future = self.fsc.callback_mcast_rpc("aabbcc", "test_func", ("a", "b", "c"), ttl=15)
        self.assertMcastCalls([mock.call({'group': None, 'args': ('test_func1', 'test_func', 'a', 'b', 'c'), 'type': 'mRPC', 'func': 'callback', 'ttl': 15})])

        # Make sure we're waiting on this particular RPC
        self.assertEqual(self.fsc._callback_futures.keys(), ["test_func1"])

        # Mock out the RPC being sent
        self.client.mock_rpc_sent(0)

        # Mock out the response to the RPC
        self.client.mock_rx("aabbcc", "test_func1", "Result")

        # Make sure we immediately resolve the future
        yield moments()
        self.assertEqual(callback_rpc_future.running(), False)

        rpc_result = yield callback_rpc_future
        self.assertEqual(rpc_result, {"aabbcc": ["Result"]})

        # Make sure we're not waiting on anything else
        self.assertEqual(self.fsc._callback_futures, {})
        self.assertEqual(self.fsc._rpc_futures, {})
        # self.assertRpcCalls([]) ???

    @gen_test
    def test_callback_mcast_rpc_with_timeout(self):
        """Test callback multicast RPC with timeouts"""
        callback_rpc_future = self.fsc.callback_mcast_rpc("aabbcc", "test_func", ("a", "b", "c"), retries=2, timeout=.05)
        self.assertMcastCalls([mock.call({'group': None, 'args': ('test_func1', 'test_func', 'a', 'b', 'c'), 'type': 'mRPC', 'func': 'callback', 'ttl': None})])
        self.client.mock_rpc_sent(0)

        # Make sure we retried (1x)
        yield sleep(.06)
        self.assertMcastCalls([mock.call({'group': None, 'args': ('test_func1', 'test_func', 'a', 'b', 'c'), 'type': 'mRPC', 'func': 'callback', 'ttl': None})])
        self.client.mock_rpc_sent(1)

        # Make sure we retried (2x)
        yield sleep(.06)
        self.assertMcastCalls([mock.call({'group': None, 'args': ('test_func1', 'test_func', 'a', 'b', 'c'), 'type': 'mRPC', 'func': 'callback', 'ttl': None})])
        self.client.mock_rpc_sent(2)

        # Make sure after the final timeout we return None
        yield sleep(.06)
        self.assertEqual(callback_rpc_future.running(), False)

        rpc_result = yield callback_rpc_future
        self.assertEqual(rpc_result, {"aabbcc": None})

        # Make sure we're not waiting on anything else
        self.assertEqual(self.fsc._callback_futures, {})
        self.assertEqual(self.fsc._rpc_futures, {})
        self.assertRpcCalls([])

        # Make sure we removed the function from the dispatcher
        self.assertNotIn("rpc/+/test_func1", self.client.subscriptions)

    @gen_test
    def test_callback_mcast_rpc_with_multiple_responses(self):
        """Test callback multicast RPC with multiple responses"""
        callback_rpc_future = self.fsc.callback_mcast_rpc(["aabbcc", "ccddee"], "test_func", ("a", "b", "c"), group=1, ttl=3)
        self.assertMcastCalls([mock.call({'group': 1, 'args': ('test_func1', 'test_func', 'a', 'b', 'c'), 'type': 'mRPC', 'func': 'callback', 'ttl': 3})])

        # Make sure we're waiting on this particular RPC
        self.assertEqual(self.fsc._callback_futures.keys(), ["test_func1"])

        # Mock out the RPC being sent
        self.client.mock_rpc_sent(0)

        # Mock out the response to the RPC
        self.client.mock_rx("ccddee", "test_func1", "Result 2")
        self.client.mock_rx("aabbcc", "test_func1", "Result 1")

        # Make sure we immediately resolve the future
        yield moments()
        self.assertEqual(callback_rpc_future.running(), False)

        rpc_result = yield callback_rpc_future
        self.assertEqual(rpc_result, {"aabbcc": ["Result 1"], "ccddee": ["Result 2"]})

        # Make sure we're not waiting on anything else
        self.assertEqual(self.fsc._callback_futures, {})
        self.assertEqual(self.fsc._rpc_futures, {})
        self.assertRpcCalls([])

    @gen_test
    def test_callback_mcast_rpc_with_named_tuple(self):
        """Test callback multicast RPC with multiple responses and named tuples"""
        example_type = namedtuple("ExampleType", ["a", "b"])
        callback_rpc_future = self.fsc.callback_mcast_rpc(["aabbcc", "ccddee"], "test_func", ("a", "b", "c"), result_type=example_type)
        self.assertMcastCalls([mock.call({'group': None, 'args': ('test_func1', 'test_func', 'a', 'b', 'c'), 'type': 'mRPC', 'func': 'callback', 'ttl': None})])

        # Make sure we're waiting on this particular RPC
        self.assertEqual(self.fsc._callback_futures.keys(), ["test_func1"])

        # Mock out the RPC being sent
        self.client.mock_rpc_sent(0)

        # Mock out the response to the RPC
        self.client.mock_rx("ccddee", "test_func1", "Result 2", 10)
        self.client.mock_rx("aabbcc", "test_func1", "Result 1", 20)

        # Make sure we immediately resolve the future
        yield moments()
        self.assertEqual(callback_rpc_future.running(), False)

        rpc_result = yield callback_rpc_future
        self.assertEqual(rpc_result, {"aabbcc": example_type("Result 1", 20), "ccddee": example_type("Result 2", 10)})
        self.assertIsInstance(rpc_result['aabbcc'], example_type)
        self.assertIsInstance(rpc_result['ccddee'], example_type)

        # Make sure we're not waiting on anything else
        self.assertEqual(self.fsc._callback_futures, {})
        self.assertEqual(self.fsc._rpc_futures, {})
        self.assertRpcCalls([])

    @gen_test
    def test_callback_dmcast_rpc_with_named_tuple(self):
        """Test callback directed multicast RPC with multiple responses and named tuples"""
        example_type = namedtuple("ExampleType", ["a", "b"])
        callback_rpc_future = self.fsc.callback_dmcast(["aabbcc", "ccddee"], "test_func", ("a", "b", "c"), result_type=example_type)
        self.assertDMcastCalls([mock.call({
            'group': 1,
            'dst': ['aabbcc', 'ccddee'],
            'args': ('test_func1', 'test_func', 'a', 'b', 'c'),
            'delay': 50,
            'func': 'callback',
            'ttl': 5,
            'type': 'dmRPC'
        })])

        # Make sure we're waiting on this particular RPC
        self.assertEqual(self.fsc._callback_futures.keys(), ["test_func1"])

        # Mock out the RPC being sent
        self.client.mock_rpc_sent(0)

        # Mock out the response to the RPC
        self.client.mock_rx("ccddee", "test_func1", "Result 2", 10)
        self.client.mock_rx("aabbcc", "test_func1", "Result 1", 20)

        # Make sure we immediately resolve the future
        yield moments()
        self.assertEqual(callback_rpc_future.running(), False)

        rpc_result = yield callback_rpc_future
        self.assertEqual(rpc_result.responses, {"aabbcc": example_type("Result 1", 20), "ccddee": example_type("Result 2", 10)})
        self.assertIsInstance(rpc_result.responses['aabbcc'], example_type)
        self.assertIsInstance(rpc_result.responses['ccddee'], example_type)

        # Make sure we're not waiting on anything else
        self.assertEqual(self.fsc._callback_futures, {})
        self.assertEqual(self.fsc._rpc_futures, {})
        self.assertRpcCalls([])

    @gen_test
    def test_callback_dmcast_rpc_timeout_with_named_tuple(self):
        """Test callback directed multicast RPC with timeouts, multiple responses, and named tuples"""
        example_type = namedtuple("ExampleType", ["a", "b"])
        callback_rpc_future = self.fsc.callback_dmcast(
            ["aabbcc", "ccddee", '112233'],
            "test_func",
            ("a", "b", "c"),
            result_type=example_type,
            timeout=.05,
            retries=1)

        # Verify the first message is sent
        self.assertDMcastCalls([mock.call({
            'group': 1,
            'dst': ['aabbcc', 'ccddee', '112233'],
            'args': ('test_func1', 'test_func', 'a', 'b', 'c'),
            'delay': 50,
            'func': 'callback',
            'ttl': 5,
            'type': 'dmRPC'
        })])

        self.client.mock_rpc_sent(0)

        # Receive the first response
        self.client.mock_rx("aabbcc", "test_func1", "Result 1", 20)

        # Experience a timeout
        yield sleep(.06)

        # Make sure we send the next message
        self.assertDMcastCalls([mock.call({
            'group': 1,
            'dst': ['ccddee', '112233'],
            'args': ('test_func2', 'test_func', 'a', 'b', 'c'),
            'delay': 50,
            'func': 'callback',
            'ttl': 5,
            'type': 'dmRPC'
        })])

        self.client.mock_rpc_sent(1)

        # Receive the response for the 2nd message
        self.client.mock_rx("112233", "test_func2", "Result 2", 10)

        # Experience a timeout
        yield sleep(.06)

        # Make sure we immediately resolve the future
        yield moments()
        self.assertEqual(callback_rpc_future.running(), False)

        rpc_result = yield callback_rpc_future
        self.assertEqual(rpc_result.responses, {"aabbcc": example_type("Result 1", 20), "112233": example_type("Result 2", 10), "ccddee": None})
        self.assertIsInstance(rpc_result.responses['aabbcc'], example_type)
        self.assertIsInstance(rpc_result.responses['112233'], example_type)

        # Make sure we're not waiting on anything else
        self.assertEqual(self.fsc._callback_futures, {})
        self.assertEqual(self.fsc._rpc_futures, {})
        self.assertRpcCalls([])

    @gen_test
    def test_callback_mcast_rpc_with_partial_responses(self):
        """Test callback multicast RPC with only some nodes responding"""
        callback_rpc_future = self.fsc.callback_mcast_rpc(["aabbcc", "ccddee"], "test_func", ("a", "b", "c"), retries=0, timeout=.1)
        self.client.mock_rpc_sent(0)

        self.client.mock_rx("ccddee", "test_func1", "Result 2")

        # Wait until we're out of time after hearing only a single response
        yield sleep(.2)

        rpc_result = yield callback_rpc_future
        self.assertEqual(rpc_result, {"aabbcc": None, "ccddee": ["Result 2"]})

        # Make sure we're not waiting on anything else
        self.assertEqual(self.fsc._callback_futures, {})
        self.assertEqual(self.fsc._rpc_futures, {})
        # self.assertRpcCalls([]) ???

        # Make sure we removed the function from the dispatcher
        self.assertNotIn("rpc/+/test_func1", self.client.subscriptions)

    @gen_test
    def test_dmcast_rpc(self):
        """Test that we can asynchronously send dmcast RPCs"""
        # Test a successful RPC
        send_rpc_future = self.fsc.dmRPC("aabbcc", "test_func", "arg1", "arg2", "arg3")
        self.assertTrue(send_rpc_future.running())
        self.assertDMcastCalls([mock.call({'group': 'test_func',
                                           'dst': 'aabbcc',
                                           'args': (),
                                           'delay': 'arg2',
                                           'func': 'arg3',
                                           'ttl': 'arg1',
                                           'type': 'dmRPC'})])

        # Make sure we save the future off
        self.assertEqual(self.fsc._rpc_futures, {0: send_rpc_future})

        # Simulate a successful RPC exiting the system
        self.client.mock_rpc_sent(0)
        self.assertEqual(self.fsc._rpc_futures, {})

        # At this point, our future has resolved, so we have a result
        rpc_result = yield send_rpc_future
        self.assertEqual(rpc_result, True)

        # send a ttl and a delay
        send_rpc_future = self.fsc.dmRPC("aabbcc", DEFAULT_MCAST_GROUP, 10, 100, "test_func", "arg1", "arg2", "arg3")
        self.assertTrue(send_rpc_future.running())
        self.assertDMcastCalls([mock.call(
            {'group': 1,
             'dst': 'aabbcc',
             'args': ('arg1', 'arg2', 'arg3'),
             'delay': 100,
             'func': 'test_func',
             'ttl': 10,
             'type': 'dmRPC'})])
        # Make sure we save the future off
        self.assertEqual(self.fsc._rpc_futures, {1: send_rpc_future})

        # Simulate a successful RPC exiting the system
        self.client.mock_rpc_sent(1)
        self.assertEqual(self.fsc._rpc_futures, {})

        # At this point, our future has resolved, so we have a result
        rpc_result = yield send_rpc_future
        self.assertEqual(rpc_result, True)

        # send a ttl and a delay AND more addresses
        send_rpc_future = self.fsc.dmRPC("aabbccddeeff001122", DEFAULT_MCAST_GROUP, 20, 200, "test_func", "arg1", "arg2", "arg3")
        self.assertTrue(send_rpc_future.running())
        self.assertDMcastCalls([mock.call(
            {'group': 1,
             'dst': 'aabbccddeeff001122',
             'args': ('arg1', 'arg2', 'arg3'),
             'delay': 200,
             'func': 'test_func',
             'ttl': 20,
             'type': 'dmRPC'})])
        # Make sure we save the future off
        self.assertEqual(self.fsc._rpc_futures, {2: send_rpc_future})

        # Simulate a successful RPC exiting the system
        self.client.mock_rpc_sent(2)  # Packet ID 2 returned True
        self.assertEqual(self.fsc._rpc_futures, {})

        # At this point, our future has resolved, so we have a result
        rpc_result = yield send_rpc_future
        self.assertEqual(rpc_result, True)

        # Try again, where the RPC fails
        send_rpc_future = self.fsc.dmRPC("aabbcc", DEFAULT_MCAST_GROUP, DEFAULT_MCAST_TTL, DEFAULT_DMCAST_DELAY, "test_func")
        self.assertTrue(send_rpc_future.running())
        self.assertDMcastCalls([mock.call({'group': 1, 'dst': 'aabbcc', 'args': (), 'delay': 50, 'func': 'test_func', 'ttl': 5, 'type': 'dmRPC'})])

        # Make sure we save the future off
        self.assertEqual(self.fsc._rpc_futures, {3: send_rpc_future})

        # Simulate an RPC exiting the system failure
        self.client.mock_rpc_sent(3, False)  # Packet ID 3 returned False
        self.assertEqual(self.fsc._rpc_futures, {})

        # At this point, our future has resolved, so we have a result
        rpc_result = yield send_rpc_future
        self.assertEqual(rpc_result, False)

    @gen_test
    def test_dmcast_rpc_with_list(self):
        """Test that the dmcast_rpc function supports a list of addresses as an argument."""
        self.fsc.dmRPC(["aabbcc", "ccddee", "112233"], "test_func", "arg1", "arg2", "arg3")
        self.assertDMcastCalls([mock.call(
            {'group': 'test_func',
             'dst': ['aabbcc', 'ccddee', '112233'],
             'args': (),
             'delay': 'arg2',
             'func': 'arg3',
             'ttl': 'arg1',
             'type': 'dmRPC'})])

    @gen_test
    def test_callback_dmcast_rpc(self):
        """Test that we can asynchronously send dmcast RPCs"""
        # Test a successful RPC with a sequence number
        callback_rpc_future = self.fsc.callback_dmcast("aabbcc", "test_func", ("arg1", "arg2", "arg3"))
        self.assertTrue(callback_rpc_future.running())
        self.assertDMcastCalls([mock.call({
            'group': 1,
            'dst': ['aabbcc'],
            'args': ('test_func1', 'test_func', 'arg1', 'arg2', 'arg3'),
            'delay': 50,
            'func': 'callback',
            'ttl': 5,
            'type': 'dmRPC'
        })])

        # Make sure we're waiting on this particular RPC
        self.assertEqual(self.fsc._callback_futures.keys(), ["test_func1"])

        # Mock out the RPC being sent
        self.client.mock_rpc_sent(0)

        # Mock out the response to the RPC
        self.client.mock_rx("aabbcc", "test_func1", "Result")

        # Make sure we immediately resolve the future
        yield moments()
        self.assertEqual(callback_rpc_future.running(), False)

        responses, unprocessed = multi_response = yield callback_rpc_future
        self.assertIsInstance(multi_response, MultiResult)
        self.assertEqual((responses, unprocessed), ({"aabbcc": ["Result"]}, []))

        # Make sure we're not waiting on anything else
        self.assertEqual(self.fsc._callback_futures, {})
        self.assertEqual(self.fsc._rpc_futures, {})
        self.assertRpcCalls([])

        # Make sure we removed the function from the dispatcher
        self.assertNotIn("rpc/+/test_func1", self.client.subscriptions)

        # send a callback expecting a sequence number
        callback_rpc_future = self.fsc.callback_dmcast("aabbcc", "seq_func", ("argA", 2, '0102ff'))
        self.assertTrue(callback_rpc_future.running())
        self.assertDMcastCalls([mock.call({
            'group': 1,
            'dst': ['aabbcc'],
            'args': ('seq_func1', 'seq_func', 'argA', 2, '0102ff'),
            'delay': 50,
            'func': 'callback',
            'ttl': 5,
            'type': 'dmRPC'
        })])

        # Make sure we're waiting on this particular RPC
        self.assertEqual(self.fsc._callback_futures.keys(), ["seq_func1"])

        # Mock out the RPC being sent
        self.client.mock_rpc_sent(1)

        # Mock out the response to the RPC
        self.client.mock_rx("aabbcc", "seq_func1", "Result")

        # Make sure we immediately resolve the future
        yield moments()
        self.assertEqual(callback_rpc_future.running(), False)

        responses, unprocessed = multi_response = yield callback_rpc_future
        self.assertIsInstance(multi_response, MultiResult)
        self.assertEqual((responses, unprocessed), ({"aabbcc": ["Result"]}, []))

        # Make sure we're not waiting on anything else
        self.assertEqual(self.fsc._callback_futures, {})
        self.assertEqual(self.fsc._rpc_futures, {})
        self.assertRpcCalls([])

    @gen_test
    def test_callback_dmcast_no_callback_name(self):
        """Test that we do not raise an error if we try to do a dmcast without a callback function expected"""
        self.fsc.callback_dmcast("aabbcc", "lsGET", ("a", "b", "c"))
        self.assertDMcastCalls([mock.call({
            'dst': ['aabbcc'],
            'group': DEFAULT_MCAST_GROUP,
            'ttl': DEFAULT_MCAST_TTL,
            'delay': DEFAULT_DMCAST_DELAY,  # dmcast arguments
            'func': 'callback',
            'args': ('lsGET1', 'lsGET', 'a', 'b', 'c'),
            'type': 'dmRPC'
        })])  # rpc arguments

    @gen_test
    def test_callback_dmcast_dmcast_delay(self):
        """Test that we can change the dmcast delay parameter"""
        self.fsc._app_seq = 0
        self.fsc.callback_dmcast("aabbcc", "lsGET", ("a", "b", "c"), callback_name="lsDAT", dmcast_delay=30)
        self.assertDMcastCalls([mock.call({
            'dst': ['aabbcc'],
            'group': DEFAULT_MCAST_GROUP,
            'ttl': DEFAULT_MCAST_TTL,
            'delay': 30,  # dmcast arguments
            'func': 'lsGET',
            'args': ('a', 'b', 'c'),
            'type': 'dmRPC'
        })])  # rpc arguments

        # When no dmcast_delay is specified, fallback to the default value
        self.fsc.callback_dmcast("aabbcc", "lsGET", ("a", "b", "c"), callback_name="lsDAT")
        self.assertDMcastCalls([mock.call({
            'dst': ['aabbcc'],
            'group': DEFAULT_MCAST_GROUP,
            'ttl': DEFAULT_MCAST_TTL,
            'delay': DEFAULT_DMCAST_DELAY,  # dmcast arguments
            'func': 'lsGET',
            'args': ('a', 'b', 'c'),
            'type': 'dmRPC'
        })])  # rpc arguments

    @gen_test
    def test_callback_dmcast_no_arguments(self):
        """Test that we can send dmcast RPC messages with the correct nodes included and no arguments"""
        callback_rpc_future = self.fsc.callback_dmcast("aabbcc", "lsGET", callback_name="lsDAT")
        self.assertDMcastCalls([mock.call({
            'dst': ['aabbcc'],
            'group': DEFAULT_MCAST_GROUP,
            'ttl': DEFAULT_MCAST_TTL,
            'delay': DEFAULT_DMCAST_DELAY,  # dmcast arguments
            'func': 'lsGET',
            'args': (),
            'type': 'dmRPC'
        })])  # rpc arguments
        self.client.mock_rpc_sent(0)

        # Mock out the response to the RPC
        self.client.mock_rx("aabbcc", "lsDAT", "payload")

        rpc_result, unprocessed = yield callback_rpc_future
        self.assertEqual(rpc_result, {"aabbcc": ["payload"]})
        self.assertEqual(unprocessed, [])

    @gen_test
    def test_callback_dmcast_no_response_arguments(self):
        """Test that we can send dmcast RPC messages with the correct nodes included and the response contains no arguments"""
        callback_rpc_future = self.fsc.callback_dmcast("aabbcc", "lsGET", callback_name="lsDAT")
        self.assertDMcastCalls([mock.call({
            'dst': ['aabbcc'],
            'group': DEFAULT_MCAST_GROUP,
            'ttl': DEFAULT_MCAST_TTL,
            'delay': DEFAULT_DMCAST_DELAY,  # dmcast arguments
            'func': 'lsGET',
            'args': (),
            'type': 'dmRPC'
        })])  # rpc arguments
        self.client.mock_rpc_sent(0)

        # Mock out the response to the RPC
        self.client.mock_rx("aabbcc", "lsDAT")

        rpc_result, unprocessed = yield callback_rpc_future
        self.assertEqual(rpc_result, {"aabbcc": []})
        self.assertEqual(unprocessed, [])

    @gen_test
    def test_callback_dmcast_multiple_nodes(self):
        """Test that we can send dmcast RPC messages with multiple nodes"""
        callback_rpc_future = self.fsc.callback_dmcast(["aabbcc", "ddeeff"], "lsGET", ("a", "b", "c"), callback_name="lsDAT")
        self.assertDMcastCalls([mock.call({
            'dst': ['aabbcc', 'ddeeff'],
            'group': DEFAULT_MCAST_GROUP,
            'ttl': DEFAULT_MCAST_TTL,
            'delay': DEFAULT_DMCAST_DELAY,  # dmcast arguments
            'func': 'lsGET',
            'args': ('a', 'b', 'c'),
            'type': 'dmRPC'
        })])  # rpc arguments

        # Make sure we're waiting on this particular RPC
        self.assertEqual(self.fsc._callback_futures.keys(), ["lsDAT"])

        # Mock out the RPC being sent
        self.client.mock_rpc_sent(0)

        # Mock out the response to the RPC
        self.client.mock_rx("aabbcc", "lsDAT", "payload 1")
        self.client.mock_rx("ddeeff", "lsDAT", "payload 2")

        # Make sure we immediately resolve the future
        yield moments()
        self.assertEqual(callback_rpc_future.running(), False)

        x = yield callback_rpc_future
        rpc_result, unprocessed = x
        self.assertEqual(rpc_result, {"aabbcc": ["payload 1"], "ddeeff": ["payload 2"]})
        self.assertEqual(unprocessed, [])

        # Make sure we're not waiting on anything else
        self.assertEqual(self.fsc._callback_futures, {})
        self.assertEqual(self.fsc._rpc_futures, {})
        self.assertRpcCalls([])

    @gen_test
    def test_callback_dmcast_node_limit(self):
        """Test that we can send dmcast RPC messages with the correct node limit"""
        callback_rpc_future = self.fsc.callback_dmcast(["aabbcc", "ddeeff", "eeffaa", "112233", "445566"],
                                                       "lsGET",
                                                       ("a", "b", "c"),
                                                       callback_name="lsDAT",
                                                       node_limit=2,
                                                       chunk_results=True)
        self.assertDMcastCalls([mock.call({
            'dst': ['aabbcc', 'ddeeff'],
            'group': DEFAULT_MCAST_GROUP,
            'ttl': DEFAULT_MCAST_TTL,
            'delay': DEFAULT_DMCAST_DELAY,  # dmcast arguments
            'func': 'lsGET',
            'args': ('a', 'b', 'c'),
            'type': 'dmRPC'
        })])  # rpc arguments

        # Make sure we're waiting on this particular RPC
        self.assertEqual(self.fsc._callback_futures.keys(), ["lsDAT"])

        # Mock out the RPC being sent
        self.client.mock_rpc_sent(0)

        # Mock out the response to the RPC
        self.client.mock_rx("aabbcc", "lsDAT", "payload 1")
        self.client.mock_rx("ddeeff", "lsDAT", "payload 2")

        # Make sure we immediately resolve the future
        yield moments()
        self.assertEqual(callback_rpc_future.running(), False)

        rpc_result, unprocessed = multi_response = yield callback_rpc_future
        self.assertEqual(rpc_result, {"aabbcc": ["payload 1"], "ddeeff": ["payload 2"]})
        self.assertEqual(unprocessed, ["eeffaa", "112233", "445566"])
        self.assertEqual(multi_response.chunk_responses, rpc_result)

        # Make sure we're not waiting on anything else
        self.assertEqual(self.fsc._callback_futures, {})
        self.assertEqual(self.fsc._rpc_futures, {})
        self.assertRpcCalls([])

    @gen_test
    def test_dmcast_retries(self):
        """Test that retries work for directed multicast"""
        callback_rpc_future = self.fsc.callback_dmcast(["aabbcc", "ddeeff"], "lsGET", ("a", "b", "c"), callback_name="lsDAT", retries=2, timeout=.05)

        # Make sure we're waiting on this particular RPC
        self.assertEqual(self.fsc._callback_futures.keys(), ["lsDAT"])

        yield sleep(.06)

        # Mock out the RPC being sent
        self.client.mock_rpc_sent(0)

        yield sleep(.06)

        # Mock out the RPC being sent
        self.client.mock_rpc_sent(1)

        yield sleep(.06)

        # Mock out the RPC being sent
        self.client.mock_rpc_sent(2)

        # Wait for retries
        yield sleep(.06)

        self.assertDMcastCalls([mock.call({
            'group': DEFAULT_MCAST_GROUP,
            'dst': ['aabbcc', 'ddeeff'],
            'args': ('a', 'b', 'c'),
            'delay': DEFAULT_DMCAST_DELAY,
            'func': 'lsGET',
            'ttl': DEFAULT_MCAST_TTL,
            'type': 'dmRPC'
        }), mock.call({
            'group': DEFAULT_MCAST_GROUP,
            'dst': ['aabbcc', 'ddeeff'],
            'args': ('a', 'b', 'c'),
            'delay': DEFAULT_DMCAST_DELAY,
            'func': 'lsGET',
            'ttl': DEFAULT_MCAST_TTL,
            'type': 'dmRPC'
        }), mock.call({
            'group': DEFAULT_MCAST_GROUP,
            'dst': ['aabbcc', 'ddeeff'],
            'args': ('a', 'b', 'c'),
            'delay': DEFAULT_DMCAST_DELAY,
            'func': 'lsGET',
            'ttl': DEFAULT_MCAST_TTL,
            'type': 'dmRPC'
        })])

        # Make sure after the final timeout we return None
        yield sleep(.06)
        self.assertEqual(callback_rpc_future.running(), False)

        rpc_result, unprocessed = yield callback_rpc_future
        self.assertEqual(rpc_result, {"aabbcc": None, "ddeeff": None})
        self.assertEqual(unprocessed, [])

        # Make sure we're not waiting on anything else
        self.assertEqual(self.fsc._callback_futures, {})
        self.assertEqual(self.fsc._rpc_futures, {})
        self.assertRpcCalls([])

    @gen_test
    def test_callback_dmcast_retries_remove_success_nodes(self):
        """Test that dmcast RPC retries do not resend nodes we've already heard from"""
        callback_rpc_future = self.fsc.callback_dmcast(["aabbcc", "ddeeff"], "lsGET", ("a", "b", "c"), callback_name="lsDAT", retries=2, timeout=.05)
        self.assertDMcastCalls([mock.call(
            {'group': 1,
             'dst': ['aabbcc', 'ddeeff'],
             'args': ('a', 'b', 'c'),
             'delay': 50,
             'func': 'lsGET',
             'ttl': 5,
             'type': 'dmRPC'})])

        # Make sure we're waiting on this particular RPC
        self.assertEqual(self.fsc._callback_futures.keys(), ["lsDAT"])

        yield sleep(.06)

        # Mock out the RPC being sent
        self.client.mock_rpc_sent(0)

        # Hear a success message from 1 of the nodes
        self.client.mock_rx("aabbcc", "lsDAT", "payload 1")

        # Make sure we retried (1x)
        yield sleep(.06)
        self.assertDMcastCalls([mock.call({'group': 1, 'dst': ['ddeeff'], 'args': ('a', 'b', 'c'), 'delay': 50, 'func': 'lsGET', 'ttl': 5, 'type': 'dmRPC'})])

        # Mock out the RPC being sent
        self.client.mock_rpc_sent(1)

        # Make sure we retried (2x)
        yield sleep(.06)
        self.assertDMcastCalls([mock.call({'group': 1, 'dst': ['ddeeff'], 'args': ('a', 'b', 'c'), 'delay': 50, 'func': 'lsGET', 'ttl': 5, 'type': 'dmRPC'})])

        # Mock out the RPC being sent
        self.client.mock_rpc_sent(2)

        # Make sure after the final timeout we return None
        yield sleep(.06)
        self.assertEqual(callback_rpc_future.running(), False)

        rpc_result, unprocessed = yield callback_rpc_future
        self.assertEqual(rpc_result, {"aabbcc": ["payload 1", ], "ddeeff": None})
        self.assertEqual(unprocessed, [])

        # Make sure we're not waiting on anything else
        self.assertEqual(self.fsc._callback_futures, {})
        self.assertEqual(self.fsc._rpc_futures, {})
        self.assertRpcCalls([])

    @gen_test
    def test_unprocessed_nodes(self):
        """Test that we only fit as many nodes as we can into a directed multicast and return a list of unprocessed nodes"""
        arg_length = 70  # At this argument length, we will only send 3 addresses at a time
        self.assertEqual(get_max_addrs(get_packet_overhead("lsGET", '', 1, 5, 5, "f" * arg_length)), 3)

        addrs = ["aabbc0", "aabbc1", "aabbc2", "aabbc3", "aabbc4"]
        callback_rpc_future = self.fsc.callback_dmcast(addrs, "lsGET", ("f" * arg_length, ), callback_name="lsDAT", chunk_results=True)

        self.assertDMcastCalls([mock.call(
            {'group': 1,
             'dst': ['aabbc0', 'aabbc1', 'aabbc2'],
             'args': ('f' * arg_length, ),
             'delay': 50,
             'func': 'lsGET',
             'ttl': 5,
             'type': 'dmRPC'})])

        # Make sure we're waiting on this particular RPC
        self.assertEqual(self.fsc._callback_futures.keys(), ["lsDAT"])

        # Mock out the RPC being sent
        self.client.mock_rpc_sent(0)

        # Mock out the responses to the RPC
        self.client.mock_rx("aabbc0", "lsDAT", "payload 0")
        self.client.mock_rx("aabbc1", "lsDAT", "payload 1")
        self.client.mock_rx("aabbc2", "lsDAT", "payload 2")

        # Make sure we immediately resolve the future
        yield moments()
        self.assertEqual(callback_rpc_future.running(), False)

        rpc_result, unprocessed = multi_response = yield callback_rpc_future
        self.assertEqual(rpc_result, {"aabbc0": ["payload 0", ], "aabbc1": ["payload 1", ], "aabbc2": ["payload 2", ]})
        self.assertEqual(unprocessed, ["aabbc3", "aabbc4"])
        self.assertEqual(rpc_result, multi_response.chunk_responses)

        # Make sure we're not waiting on anything else
        self.assertEqual(self.fsc._callback_futures, {})
        self.assertEqual(self.fsc._rpc_futures, {})
        self.assertRpcCalls([])

        # Make sure we removed the function from the dispatcher
        self.assertNotIn("rpc/+/lsDAT", self.client.subscriptions)

    @gen_test
    def test_immediately_resolve_rpcs(self):
        """Test that we can optionally resolve RPCs without waiting on HOOK_RPC_SENT"""
        self.fsc.immediately_resolve_rpcs()

        # Test RPC
        send_rpc_future = self.fsc.uRPC("aabbcc", "test_func", "arg1", "arg2", "arg3")
        self.assertFalse(send_rpc_future.running())
        self.assertRpcCalls([mock.call({'dst': 'aabbcc', 'args': ('arg1', 'arg2', 'arg3'), 'type': 'uRPC', 'func': 'test_func'})])
        result = yield send_rpc_future
        self.assertTrue(result)

        # Test Multicast RPC
        send_rpc_future = self.fsc.mRPC("test_func", "arg1", "arg2", "arg3")
        self.assertFalse(send_rpc_future.running())
        self.assertMcastCalls([mock.call({'group': 'test_func', 'args': ('arg3', ), 'type': 'mRPC', 'func': 'arg2', 'ttl': 'arg1'})])
        result = yield send_rpc_future
        self.assertTrue(result)

        # Test Dmcast RPC
        send_rpc_future = self.fsc.dmRPC("aabbcc", "test_func", "arg1", "arg2", "arg3")
        self.assertFalse(send_rpc_future.running())
        self.assertDMcastCalls([mock.call({'group': 'test_func',
                                           'dst': 'aabbcc',
                                           'args': (),
                                           'delay': 'arg2',
                                           'func': 'arg3',
                                           'ttl': 'arg1',
                                           'type': 'dmRPC'})])
        result = yield send_rpc_future
        self.assertTrue(result)

    @gen_test
    def test_next_message_future(self):
        """Test that we can wait on the next message future"""
        self.fsc.immediately_resolve_rpcs()

        # Test RPC
        next_message_future = self.fsc.next_message_sent
        send_rpc_future = self.fsc.uRPC("aabbcc", "test_func", "arg1", "arg2", "arg3")
        self.assertFalse(send_rpc_future.running())
        self.assertRpcCalls([mock.call({'dst': 'aabbcc', 'args': ('arg1', 'arg2', 'arg3'), 'type': 'uRPC', 'func': 'test_func'})])
        result = yield next_message_future
        self.assertTrue(result)

        # Test Multicast RPC
        next_message_future = self.fsc.next_message_sent
        send_rpc_future = self.fsc.mRPC("test_func", "arg1", "arg2", "arg3")
        self.assertFalse(send_rpc_future.running())
        self.assertMcastCalls([mock.call({'group': 'test_func', 'args': ('arg3', ), 'type': 'mRPC', 'func': 'arg2', 'ttl': 'arg1'})])
        result = yield next_message_future
        self.assertTrue(result)

        # Test Dmcast RPC
        next_message_future = self.fsc.next_message_sent
        send_rpc_future = self.fsc.dmRPC("aabbcc", "test_func", "arg1", "arg2", "arg3")
        self.assertFalse(send_rpc_future.running())
        self.assertDMcastCalls([mock.call({'group': 'test_func',
                                           'dst': 'aabbcc',
                                           'args': (),
                                           'delay': 'arg2',
                                           'func': 'arg3',
                                           'ttl': 'arg1',
                                           'type': 'dmRPC'})])

        result = yield next_message_future
        self.assertTrue(result)

    @gen_test
    def test_disable_immediately_resolve_rpcs(self):
        """Test that we can optionally resolve RPCs without waiting on HOOK_RPC_SENT and then disable the feature"""
        self.client = get_mock_client()
        self.fsc = FutureClient(self.client)

        def mock_tx(mid):
            self.fsc._on_publish(None, None, mid)
            self.fsc._message_sent()

        self.fsc.mock_tx = mock_tx
        self.client.mock_rpc_sent = mock_tx

        # Enable immediate resolve of RPCs
        self.fsc.immediately_resolve_rpcs(True)
        self.assertTrue(self.fsc._immediately_resolve_rpcs)

        # Test RPC
        send_rpc_future = self.fsc.uRPC("aabbcc", "test_func", "arg1", "arg2", "arg3")
        self.assertFalse(send_rpc_future.running())
        self.assertRpcCalls([mock.call({'dst': 'aabbcc', 'args': ('arg1', 'arg2', 'arg3'), 'type': 'uRPC', 'func': 'test_func'})])
        result = yield send_rpc_future
        self.assertTrue(result)

        # Disable the feature and make sure we still need to wait
        self.fsc.immediately_resolve_rpcs(False)
        self.assertFalse(self.fsc._immediately_resolve_rpcs)

        send_rpc_future = self.fsc.uRPC("aabbcc", "test_func", "arg1", "arg2", "arg3")
        self.assertTrue(send_rpc_future.running())
        self.assertRpcCalls([mock.call({'dst': 'aabbcc', 'args': ('arg1', 'arg2', 'arg3'), 'type': 'uRPC', 'func': 'test_func'})])
        self.client.mock_rpc_sent(1)
        result = yield send_rpc_future
        self.assertTrue(result)

    def test_multi_result_success_and_failed(self):
        """Test that the MultiResult can select only the successful or failed nodes."""
        result = MultiResult(responses={
            "aabbc0": ("Test", ),
            "aabbc1": (),
            "aabbc2": (None, ),
            "badba1": None,
            "badba2": None,
        },
                             unprocessed=['dddd01', 'dddd02'])

        # Check for the successful nodes
        self.assertEqual(result.successful, {"aabbc0": ("Test", ), "aabbc1": (), "aabbc2": (None, ), })

        # Check for the failed nodes
        self.assertIn("badba1", result.failed)
        self.assertIn("badba2", result.failed)

    def test_sequence_numbers_wrap(self):
        """Test that the sequence numbers wrap at 255."""
        self.fsc._app_seq = 1  # This starts at a random number, so fix it for this test
        self.assertEqual(self.fsc.get_next_seq(), 2)
        self.assertEqual(self.fsc._app_seq, 2)

        self.assertEqual(self.fsc.get_next_seq(), 3)
        self.assertEqual(self.fsc._app_seq, 3)

        # Make sure we wrap around
        self.fsc._app_seq = 255
        self.assertEqual(self.fsc.get_next_seq(), 1)
        self.assertEqual(self.fsc._app_seq, 1)

    def test_sort_by_attempts_remaining(self):
        """Test that we can sort items by the number of attempts remaining."""
        original_retry_counts = {"000001": 2, "000002": 1, "000003": 1, "000004": 3, "000005": 3, "000006": 2, "000007": 1, }
        sorted_values = self.fsc._sort_by_attempts_remaining(original_retry_counts)

        # Make sure every value is persent in the sorted list
        self.assertEqual(set(original_retry_counts.keys()), set(sorted_values))

        # Make sure they are in the correct order
        previous_value = sorted_values[0]
        for value in sorted_values[1:]:
            self.assertLessEqual(original_retry_counts[value], original_retry_counts[previous_value])
            previous_value = value

    @gen_test
    def test_dmcast_with_retries_retrieve_all(self):
        """Test that we can retrieve every piece of data with retries during a dmcast.

        Assumes that only 2 addresses fit in each packet
        targets = A B C D E F 0
        retries = 2 (attempts = 3)
                                        Timeouts remaining
        Message Out     Responses       A B C D E F 0
        A B             A B             X X 3 3 3 3 3
        C D             D               X X 2 X 3 3 3
        E F                             X X 2 X 2 2 3
        0 C             0 C             X X X X 2 2 X
        E F             E               X X X X X 1 X
        F                               X X X X X 0 X
        """
        targets = ['10000a', '10000b', '10000c', '10000d', '10000e', '10000f', '100000']

        callback_rpc_future = self.fsc.callback_dmcast(targets, "test_func", ("a", "b", "c"), chunk_results=False, node_limit=2, timeout=.05, retries=2)

        # First message goes to A and B
        self.assertDMcastCalls([mock.call({
            'group': 1,
            'dst': ['10000a', '10000b'],
            'args': ('test_func1', 'test_func', 'a', 'b', 'c'),
            'delay': 50,
            'func': 'callback',
            'ttl': 5,
            'type': 'dmRPC'
        })])

        self.client.mock_rpc_sent(0)

        # Receive responses from A and B
        self.client.mock_rx("10000a", "test_func1", "Result A")
        self.client.mock_rx("10000b", "test_func1", "Result B")
        yield moments()

        # The next message goes to C and D
        self.assertDMcastCalls([mock.call({
            'group': 1,
            'dst': ['10000c', '10000d'],
            'args': ('test_func2', 'test_func', 'a', 'b', 'c'),
            'delay': 50,
            'func': 'callback',
            'ttl': 5,
            'type': 'dmRPC'
        })])
        self.client.mock_rpc_sent(1)

        # Receive responses from D
        self.client.mock_rx("10000d", "test_func2", "Result D")
        yield sleep(.06)  # Wait for our timeout

        # The next message goes to E and F
        self.assertDMcastCalls([mock.call({
            'group': 1,
            'dst': ['10000e', '10000f'],
            'args': ('test_func3', 'test_func', 'a', 'b', 'c'),
            'delay': 50,
            'func': 'callback',
            'ttl': 5,
            'type': 'dmRPC'
        })])
        self.client.mock_rpc_sent(2)

        # Receive no responses
        yield sleep(.06)  # Wait for our timeout

        # The next message goes to 0 and C
        self.assertDMcastCalls([mock.call({
            'group': 1,
            'dst': ['100000', '10000c'],
            'args': ('test_func4', 'test_func', 'a', 'b', 'c'),
            'delay': 50,
            'func': 'callback',
            'ttl': 5,
            'type': 'dmRPC'
        })])
        self.client.mock_rpc_sent(3)

        # Receive a response from 0 and C
        self.client.mock_rx("100000", "test_func4", "Result 0")
        self.client.mock_rx("10000c", "test_func4", "Result C")
        yield moments()

        # The next message goes to E and F
        self.assertDMcastCalls([mock.call({
            'group': 1,
            'dst': ['10000e', '10000f'],
            'args': ('test_func5', 'test_func', 'a', 'b', 'c'),
            'delay': 50,
            'func': 'callback',
            'ttl': 5,
            'type': 'dmRPC'
        })])
        self.client.mock_rpc_sent(4)

        # Receive a response from E
        self.client.mock_rx("10000e", "test_func5", "Result E")
        yield sleep(.06)

        # The last message goes to F
        self.assertDMcastCalls([mock.call(
            {'group': 1,
             'dst': ['10000f'],
             'args': ('test_func6', 'test_func', 'a', 'b', 'c'),
             'delay': 50,
             'func': 'callback',
             'ttl': 5,
             'type': 'dmRPC'})])

        self.client.mock_rpc_sent(5)

        # Receive no response
        yield sleep(.06)

        # Verify our final result
        final_result = yield callback_rpc_future
        self.assertIsInstance(final_result, MultiResult)
        self.assertEqual(final_result.responses, {
            '10000a': ['Result A'],
            '10000b': ['Result B'],
            '10000c': ['Result C'],
            '10000d': ['Result D'],
            '10000e': ['Result E'],
            '10000f': None,
            '100000': ['Result 0'],
        })
        self.assertEqual(final_result.responses, final_result.chunk_responses)

    @gen_test
    def test_dmcast_with_retries_chunks(self):
        """Test that we can retrieve every piece of data with retries during a dmcast in chunks

        Assumes that only 2 addresses fit in each packet
        targets = A B C D E F 0
        retries = 2
                                        Timeouts remaining
        Message Out     Responses       A B C D E F 0
        A B             A B             X X 3 3 3 3 3
        C D             D               X X 2 X 3 3 3
        E F                             X X 2 X 2 2 3
        0 C             0 C             X X X X 2 2 X
        E F             E               X X X X X 1 X
        F                               X X X X X 0 X
        """
        targets = ['10000a', '10000b', '10000c', '10000d', '10000e', '10000f', '100000']

        # Calling with chunk results set to True!
        callback_rpc_future = self.fsc.callback_dmcast(targets, "test_func", ("a", "b", "c"), chunk_results=True, node_limit=2, timeout=.05, retries=2)

        # First message goes to A and B
        self.assertDMcastCalls([mock.call({
            'group': 1,
            'dst': ['10000a', '10000b'],
            'args': ('test_func1', 'test_func', 'a', 'b', 'c'),
            'delay': 50,
            'func': 'callback',
            'ttl': 5,
            'type': 'dmRPC'
        })])
        self.client.mock_rpc_sent(0)

        # Receive responses from A and B
        self.client.mock_rx("10000a", "test_func1", "Result A")
        self.client.mock_rx("10000b", "test_func1", "Result B")

        result = yield callback_rpc_future
        self.assertIsInstance(result, MultiResult)
        self.assertEqual(result.chunk_responses, {"10000a": ["Result A"], "10000b": ["Result B"], })
        self.assertEqual(result.responses, {"10000a": ["Result A"], "10000b": ["Result B"], })

        # Send the retry
        callback_rpc_future = result.retry()

        # The next message goes to C and D
        self.assertDMcastCalls([mock.call({
            'group': 1,
            'dst': ['10000c', '10000d'],
            'args': ('test_func2', 'test_func', 'a', 'b', 'c'),
            'delay': 50,
            'func': 'callback',
            'ttl': 5,
            'type': 'dmRPC'
        })])

        self.client.mock_rpc_sent(1)

        # Receive responses from D
        self.client.mock_rx("10000d", "test_func2", "Result D")
        yield sleep(.06)  # Wait for our timeout

        result = yield callback_rpc_future
        self.assertIsInstance(result, MultiResult)
        self.assertEqual(result.chunk_responses, {"10000c": None, "10000d": ["Result D", ], })
        self.assertEqual(result.responses, {"10000a": ["Result A", ], "10000b": ["Result B", ], "10000c": None, "10000d": ["Result D", ], })

        # Send the retry
        callback_rpc_future = result.retry()

        # The next message goes to E and F
        self.assertDMcastCalls([mock.call({
            'group': 1,
            'dst': ['10000e', '10000f'],
            'args': ('test_func3', 'test_func', 'a', 'b', 'c'),
            'delay': 50,
            'func': 'callback',
            'ttl': 5,
            'type': 'dmRPC'
        })])

        self.client.mock_rpc_sent(2)

        # Receive no responses
        yield sleep(.06)  # Wait for our timeout

        result = yield callback_rpc_future
        self.assertIsInstance(result, MultiResult)
        self.assertEqual(result.chunk_responses, {"10000e": None, "10000f": None, })
        self.assertEqual(result.responses, {
            "10000a": ["Result A", ],
            "10000b": ["Result B", ],
            "10000c": None,
            "10000d": ["Result D", ],
            "10000e": None,
            "10000f": None,
        })

        # Send the retry
        callback_rpc_future = result.retry()

        # The next message goes to 0 and C
        self.assertDMcastCalls([mock.call({
            'group': 1,
            'dst': ['100000', '10000c'],
            'args': ('test_func4', 'test_func', 'a', 'b', 'c'),
            'delay': 50,
            'func': 'callback',
            'ttl': 5,
            'type': 'dmRPC'
        })])

        self.client.mock_rpc_sent(3)

        # Receive a response from 0 and C
        self.client.mock_rx("100000", "test_func4", "Result 0")
        self.client.mock_rx("10000c", "test_func4", "Result C")
        yield moments()

        result = yield callback_rpc_future
        self.assertIsInstance(result, MultiResult)
        self.assertEqual(result.chunk_responses, {"100000": ["Result 0", ], "10000c": ["Result C", ], })
        self.assertEqual(result.responses, {
            "10000a": ["Result A", ],
            "10000b": ["Result B", ],
            "10000c": ["Result C", ],
            "10000d": ["Result D", ],
            "10000e": None,
            "10000f": None,
            "100000": ["Result 0", ],
        })

        # Send the retry
        callback_rpc_future = result.retry()

        # The next message goes to E and F
        self.assertDMcastCalls([mock.call({
            'group': 1,
            'dst': ['10000e', '10000f'],
            'args': ('test_func5', 'test_func', 'a', 'b', 'c'),
            'delay': 50,
            'func': 'callback',
            'ttl': 5,
            'type': 'dmRPC'
        })])

        self.client.mock_rpc_sent(4)

        # Receive a response from E
        self.client.mock_rx("10000e", "test_func5", "Result E")
        yield sleep(.06)

        result = yield callback_rpc_future
        self.assertIsInstance(result, MultiResult)
        self.assertEqual(result.chunk_responses, {"10000e": ["Result E", ], "10000f": None, })
        self.assertEqual(result.responses, {
            "10000a": ["Result A", ],
            "10000b": ["Result B", ],
            "10000c": ["Result C", ],
            "10000d": ["Result D", ],
            "10000e": ["Result E", ],
            "10000f": None,
            "100000": ["Result 0", ],
        })

        # Send the retry
        callback_rpc_future = result.retry()

        # The last message goes to F
        self.assertDMcastCalls([mock.call({
            'group': 1,
            'dst': ['10000f'],
            'args': ('test_func6', 'test_func', 'a', 'b', 'c'),
            'delay': 50,
            'func': 'callback',
            'ttl': 5,
            'type': 'dmRPC'
        })])

        self.client.mock_rpc_sent(5)

        # Receive no response
        yield sleep(.06)

        result = yield callback_rpc_future
        self.assertIsInstance(result, MultiResult)
        self.assertEqual(result.chunk_responses, {"10000f": None, })
        self.assertEqual(result.responses, {
            "10000a": ["Result A", ],
            "10000b": ["Result B", ],
            "10000c": ["Result C", ],
            "10000d": ["Result D", ],
            "10000e": ["Result E", ],
            "10000f": None,
            "100000": ["Result 0", ],
        })

        # Send the last retry
        callback_rpc_future = result.retry()
        self.assertDMcastCalls([])  # Make sure we didn't actually send any messages, we're done at this point

        # Verify our final result - The chunking method returns None on retry when we're finished
        final_result = yield callback_rpc_future
        self.assertIsNone(final_result)

    @gen_test
    def test_invalid_response_type_rpc(self):
        """Test that an RPC responds with None if the response RPC has too many or too few arguments."""
        example_type = namedtuple("ExampleType", ["a", "b", "c"])

        # Mock out the RPC being sent and responded with a short response
        callback_rpc_future = self.fsc.callback_rpc("aabbcc", "test_func", result_type=example_type)
        self.client.mock_rpc_sent(0)
        self.client.mock_rx("aabbcc", "test_func1", "Result1", "Result2")

        rpc_result = yield callback_rpc_future
        self.assertIsNone(rpc_result)

        # Mock out the RPC being sent and responded with a long response
        callback_rpc_future = self.fsc.callback_rpc("aabbcc", "test_func", result_type=example_type)
        self.client.mock_rpc_sent(1)
        self.client.mock_rx("aabbcc", "test_func2", "Result1", "Result2", "Result3", "Result4")

        rpc_result = yield callback_rpc_future
        self.assertIsNone(rpc_result)

    @gen_test
    def test_invalid_response_type_dmcast_rpc(self):
        """Test that a dmcast RPC responds with None if the response RPC has too many or too few arguments."""
        example_type = namedtuple("ExampleType", ["a", "b", "c"])

        # Mock out the RPC being sent and responded with a short response
        callback_rpc_future = self.fsc.callback_dmcast("aabbcc", "test_func", result_type=example_type)
        self.client.mock_rpc_sent(0)
        self.client.mock_rx("aabbcc", "test_func1", "Result1", "Result2")

        # Make sure the multi response returned None for the responses, but still stored the invalid response
        multi_response = yield callback_rpc_future
        self.assertIsInstance(multi_response, MultiResult)
        self.assertEqual(multi_response.responses, {"aabbcc": None})
        self.assertEqual(multi_response.invalid_responses, {"aabbcc": ["Result1", "Result2"]})

        # Try again, with an RPC that is too long
        callback_rpc_future = self.fsc.callback_dmcast("aabbcc", "test_func", result_type=example_type)
        self.client.mock_rpc_sent(1)
        self.client.mock_rx("aabbcc", "test_func2", "Result1", "Result2", "Result3", "Result4")

        # Make sure the multi response returned None for the responses, but still stored the invalid response
        multi_response = yield callback_rpc_future
        self.assertIsInstance(multi_response, MultiResult)
        self.assertEqual(multi_response.responses, {"aabbcc": None})
        self.assertEqual(multi_response.invalid_responses, {"aabbcc": ["Result1", "Result2", "Result3", "Result4"]})

    @gen_test
    def test_invalid_response_type_timeout_dmcast_rpc(self):
        """Test that a dmcast RPC responds with None if the response RPC has too many or too few arguments with a timeout."""
        example_type = namedtuple("ExampleType", ["a", "b", "c"])

        # Mock out the RPC being sent and responded with a short response
        callback_rpc_future = self.fsc.callback_dmcast(["aabbcc", "ccddee"], "test_func", result_type=example_type, timeout=.05, retries=0)
        self.client.mock_rpc_sent(0)
        self.client.mock_rx("aabbcc", "test_func1", "Result1", "Result2")

        # Make sure the multi response returned None for the responses, but still stored the invalid response
        multi_response = yield callback_rpc_future
        self.assertIsInstance(multi_response, MultiResult)
        self.assertEqual(multi_response.responses, {"aabbcc": None, "ccddee": None})
        self.assertEqual(multi_response.invalid_responses, {"aabbcc": ["Result1", "Result2"]})

        # Try again, with an RPC that is too long
        callback_rpc_future = self.fsc.callback_dmcast(["aabbcc", "ccddee"], "test_func", result_type=example_type, timeout=.05, retries=0)
        self.client.mock_rpc_sent(1)
        self.client.mock_rx("aabbcc", "test_func2", "Result1", "Result2", "Result3", "Result4")

        # Make sure the multi response returned None for the responses, but still stored the invalid response
        multi_response = yield callback_rpc_future
        self.assertIsInstance(multi_response, MultiResult)
        self.assertEqual(multi_response.responses, {"aabbcc": None, "ccddee": None})
        self.assertEqual(multi_response.invalid_responses, {"aabbcc": ["Result1", "Result2", "Result3", "Result4"]})

    @gen_test
    def test_invalid_response_type_dmcast_rpc_no_retries(self):
        """Test that a dmcast RPC with an invalid response does not have retries."""
        example_type = namedtuple("ExampleType", ["a", "b", "c"])

        # Mock out the RPC being sent and responded with a short response
        callback_rpc_future = self.fsc.callback_dmcast("aabbcc", "test_func", result_type=example_type, retries=10)
        self.client.mock_rpc_sent(0)
        self.client.mock_rx("aabbcc", "test_func1", "Result1", "Result2")

        # Make sure the multi response returned None for the responses, but still stored the invalid response
        multi_response = yield callback_rpc_future
        self.assertIsInstance(multi_response, MultiResult)
        self.assertEqual(multi_response.responses, {"aabbcc": None})
        self.assertEqual(multi_response.invalid_responses, {"aabbcc": ["Result1", "Result2"]})
